# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import tarfile
import logging

from concurrent.futures import ThreadPoolExecutor, as_completed

import pytest

from ovirt_imageio import client
from ovirt_imageio._internal import blkhash
from ovirt_imageio._internal import config
from ovirt_imageio._internal import ipv6
from ovirt_imageio._internal import nbd
from ovirt_imageio._internal import nbdutil
from ovirt_imageio._internal import qemu_img
from ovirt_imageio._internal import qemu_nbd
from ovirt_imageio._internal import server

from ovirt_imageio._internal.extent import ZeroExtent, DirtyExtent

from . import testutil
from . import ci

log = logging.getLogger("test")

CLUSTER_SIZE = 64 * 1024
IMAGE_SIZE = 3 * CLUSTER_SIZE


@pytest.fixture(scope="module")
def srv():
    cfg = config.load(["test/conf/daemon.conf"])
    s = server.Server(cfg)
    s.start()
    yield s
    s.stop()


def prepare_transfer(srv, url, sparse=True, size=IMAGE_SIZE):
    ticket = testutil.create_ticket(
        url=url,
        size=size,
        sparse=sparse,
        ops=["read", "write"])

    srv.auth.add(ticket)

    host, port = srv.remote_service.address
    host = ipv6.quote_address(host)
    return "https://{}:{}/images/{}".format(host, port, ticket["uuid"])


class FakeProgress:

    def __init__(self):
        self.size = None
        self.updates = []

    def update(self, n):
        self.updates.append(n)


# TODO:
# - verify that upload optimized the upload using unix socket. Need a way to
#   enable only OPTIONS on the remote server.
# - verify that upload fall back to HTTPS if server does not support unix
#   socket. We don't have a way to disable unix socket currently.
# - verify that upload fall back to HTTPS if server support unix socket but is
#   not the local host. Probbly not feasble for these tests, unless we can
#   start a daemon on another host.
# - Test negative flows


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_upload_empty_sparse(tmpdir, srv, fmt):
    src = str(tmpdir.join("src"))
    qemu_img.create(src, fmt, size=IMAGE_SIZE)

    dst = str(tmpdir.join("dst"))
    with open(dst, "wb") as f:
        f.write(b"a" * IMAGE_SIZE)

    url = prepare_transfer(srv, "file://" + dst)

    client.upload(src, url, srv.config.tls.ca_file)

    # We cannot compare allocation of raw and qcow2 images since qemu-img
    # reports holes only in qcow2 images.
    qemu_img.compare(src, dst, format1=fmt, format2="raw", strict=fmt == "raw")


@pytest.mark.parametrize("fmt", [
    pytest.param(
        "raw",
        marks=pytest.mark.xfail(
            ci.is_ovirt(),
            reason="Broken in oVirt CI for unknown reason")
    ),
    "qcow2"
])
def test_upload_hole_at_start_sparse(tmpdir, srv, fmt):
    size = 3 * 1024**2
    src = str(tmpdir.join("src"))
    qemu_img.create(src, fmt, size=size)

    with qemu_nbd.open(src, fmt) as c:
        c.write(size - 1024**2, b"b" * 1024**2)
        c.flush()
        log.debug("src extents: %s", list(nbdutil.extents(c)))

    dst = str(tmpdir.join("dst"))
    with open(dst, "wb") as f:
        f.write(b"a" * size)

    url = prepare_transfer(srv, "file://" + dst, size=size)

    client.upload(src, url, srv.config.tls.ca_file)

    with qemu_nbd.open(dst, "raw", read_only=True) as c:
        log.debug("dst extents: %s", list(nbdutil.extents(c)))

    qemu_img.compare(src, dst, format1=fmt, format2="raw", strict=fmt == "raw")


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_upload_hole_at_middle_sparse(tmpdir, srv, fmt):
    size = 3 * 1024**2
    src = str(tmpdir.join("src"))
    qemu_img.create(src, fmt, size=size)

    with qemu_nbd.open(src, fmt) as c:
        c.write(0, b"b" * 1024**2)
        c.write(size - 1024**2, b"b" * 1024**2)
        c.flush()

    dst = str(tmpdir.join("dst"))
    with open(dst, "wb") as f:
        f.write(b"a" * size)

    url = prepare_transfer(srv, "file://" + dst, size=size)

    client.upload(src, url, srv.config.tls.ca_file)

    qemu_img.compare(src, dst, format1=fmt, format2="raw", strict=fmt == "raw")


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_upload_hole_at_end_sparse(tmpdir, srv, fmt):
    size = 3 * 1024**2
    src = str(tmpdir.join("src"))
    qemu_img.create(src, fmt, size=size)

    with qemu_nbd.open(src, fmt) as c:
        c.write(0, b"b" * 1024**2)
        c.flush()

    dst = str(tmpdir.join("dst"))
    with open(dst, "wb") as f:
        f.write(b"a" * size)

    url = prepare_transfer(srv, "file://" + dst, size=size)

    client.upload(src, url, srv.config.tls.ca_file)

    qemu_img.compare(src, dst, format1=fmt, format2="raw", strict=fmt == "raw")


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_upload_full_sparse(tmpdir, srv, fmt):
    src = str(tmpdir.join("src"))
    qemu_img.create(src, fmt, size=IMAGE_SIZE)

    with qemu_nbd.open(src, fmt) as c:
        c.write(0, b"b" * IMAGE_SIZE)
        c.flush()

    dst = str(tmpdir.join("dst"))
    with open(dst, "wb") as f:
        f.write(b"a" * IMAGE_SIZE)

    url = prepare_transfer(srv, "file://" + dst)

    client.upload(src, url, srv.config.tls.ca_file)

    qemu_img.compare(src, dst, strict=True)


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_upload_preallocated(tmpdir, srv, fmt):
    src = str(tmpdir.join("src"))
    qemu_img.create(src, fmt, size=IMAGE_SIZE)

    dst = str(tmpdir.join("dst"))
    with open(dst, "wb") as f:
        f.write(b"a" * IMAGE_SIZE)

    url = prepare_transfer(srv, "file://" + dst, sparse=False)

    client.upload(src, url, srv.config.tls.ca_file)

    qemu_img.compare(src, dst)
    assert os.stat(dst).st_blocks * 512 == IMAGE_SIZE


@pytest.mark.parametrize("fmt,compressed", [
    ("raw", False),
    ("qcow2", False),
    ("qcow2", True),
])
def test_upload_from_ova(tmpdir, srv, fmt, compressed):
    offset = CLUSTER_SIZE
    data = b"I can eat glass and it doesn't hurt me."

    # Create raw disk with some data.
    tmp = str(tmpdir.join("tmp"))
    with open(tmp, "wb") as f:
        f.truncate(IMAGE_SIZE)
        f.seek(offset)
        f.write(data)

    # Create source disk.
    src = str(tmpdir.join("src"))
    qemu_img.convert(tmp, src, "raw", fmt, compressed=compressed)

    # Create OVA package.
    ova = str(tmpdir.join("src.ova"))
    with tarfile.open(ova, "w") as tar:
        tar.add(src, arcname=os.path.basename(src))

    # Prepare destination file.
    dst = str(tmpdir.join("dst"))
    with open(dst, "wb") as f:
        f.truncate(IMAGE_SIZE)

    # Test uploading src from ova.
    url = prepare_transfer(srv, "file://" + dst)
    client.upload(
        ova,
        url,
        srv.config.tls.ca_file,
        member=os.path.basename(src))

    qemu_img.compare(src, dst)


@pytest.mark.parametrize("disk_is_zero", [True, False])
@pytest.mark.parametrize("src_fmt", ["raw", "qcow2"])
@pytest.mark.parametrize("dst_fmt", ["raw", "qcow2"])
def test_upload_nbd(srv, nbd_server, tmpdir, src_fmt, dst_fmt, disk_is_zero):
    size = 5 * CLUSTER_SIZE

    # Create qcow2 src image with all kinds of extents.
    src = str(tmpdir.join("src." + src_fmt))
    qemu_img.create(src, src_fmt, size=size)
    with qemu_nbd.open(src, src_fmt) as c:
        # Cluster 0: data.
        c.write(0 * CLUSTER_SIZE, b"A" * CLUSTER_SIZE)
        # Cluster 1: data, zero.
        c.write(1 * CLUSTER_SIZE, b"\0" * CLUSTER_SIZE)
        # Cluster 2: qcow2: zero cluster, raw: hole.
        c.zero(2 * CLUSTER_SIZE, CLUSTER_SIZE, punch_hole=True)
        # Cluster 3: qcow2: zero cluster, raw: allocated.
        c.zero(3 * CLUSTER_SIZE, CLUSTER_SIZE, punch_hole=False)
        # Cluster 4: unallocated.
        c.flush()
        log.debug("src extents: %s", list(nbdutil.extents(c)))

    # Create empty target image.
    dst = str(tmpdir.join("dst." + dst_fmt))
    qemu_img.create(dst, dst_fmt, size=size)

    # Upload image.

    nbd_server.image = dst
    nbd_server.fmt = dst_fmt
    nbd_server.start()

    url = prepare_transfer(srv, nbd_server.sock.url(), size=size)
    client.upload(
        src,
        url,
        srv.config.tls.ca_file,
        disk_is_zero=disk_is_zero)

    nbd_server.stop()

    with qemu_nbd.open(dst, dst_fmt, read_only=True) as c:
        log.debug("dst extents: %s", list(nbdutil.extents(c)))

    # Compare image content. We cannot compare allocation since we skip zeroes
    # and sparsify the destination image.
    qemu_img.compare(src, dst, format1=src_fmt, format2=dst_fmt, strict=False)


@pytest.mark.parametrize("disk_is_zero", [True, False])
@pytest.mark.parametrize("base_fmt", ["raw", "qcow2"])
def test_upload_nbd_shallow(srv, nbd_server, tmpdir, base_fmt, disk_is_zero):
    size = 10 * 1024**2

    # Create base image with some data in first 3 clusters.
    src_base = str(tmpdir.join("src_base." + base_fmt))
    qemu_img.create(src_base, base_fmt, size=size)
    with qemu_nbd.open(src_base, base_fmt) as c:
        c.write(0 * CLUSTER_SIZE, b"a" * CLUSTER_SIZE)
        c.write(1 * CLUSTER_SIZE, b"b" * CLUSTER_SIZE)
        c.write(2 * CLUSTER_SIZE, b"c" * CLUSTER_SIZE)
        c.flush()
        log.debug("src_base extents: %s", list(nbdutil.extents(c)))

    # Create src image with some data in second cluster and zero in third
    # cluster.
    src_top = str(tmpdir.join("src_top.qcow2"))
    qemu_img.create(
        src_top, "qcow2", backing_file=src_base, backing_format=base_fmt)
    with qemu_nbd.open(src_top, "qcow2") as c:
        c.write(1 * CLUSTER_SIZE, b"B" * CLUSTER_SIZE)
        c.zero(2 * CLUSTER_SIZE, CLUSTER_SIZE)
        c.flush()

    with qemu_nbd.open(
            src_top, "qcow2", read_only=True, backing_chain=False) as c:
        log.debug("src_top extents: %s", list(nbdutil.extents(c)))

    # Create empty destination base image.
    dst_base = str(tmpdir.join("dst_base." + base_fmt))
    qemu_img.create(dst_base, base_fmt, size=size)

    # Create empty destination top image.
    dst_top = str(tmpdir.join("dst_top.qcow2"))
    qemu_img.create(
        dst_top, "qcow2", backing_file=dst_base, backing_format=base_fmt)

    # Upload base image.

    nbd_server.image = dst_base
    nbd_server.fmt = base_fmt
    nbd_server.start()

    url = prepare_transfer(srv, nbd_server.sock.url(), size=size)
    client.upload(
        src_base,
        url,
        srv.config.tls.ca_file,
        disk_is_zero=disk_is_zero,
        backing_chain=False)

    nbd_server.stop()

    with qemu_nbd.open(dst_base, base_fmt, read_only=True) as c:
        log.debug("dst_base extents: %s", list(nbdutil.extents(c)))

    # Compare image content - must match.
    qemu_img.compare(
        src_base, dst_base, format1=base_fmt, format2=base_fmt, strict=False)

    # Comparing allocation is not possible with raw format since qemu-nbd does
    # not report allocation for raw images, so we treat unallocated areas as
    # zero area instead of a hole.
    if base_fmt == "qcow2":
        qemu_img.compare(
            src_base, dst_base, format1=base_fmt, format2=base_fmt,
            strict=True)

    # Upload top image.

    nbd_server.image = dst_top
    nbd_server.fmt = "qcow2"
    nbd_server.start()

    url = prepare_transfer(srv, nbd_server.sock.url(), size=size)
    client.upload(
        src_top,
        url,
        srv.config.tls.ca_file,
        disk_is_zero=disk_is_zero,
        backing_chain=False)

    nbd_server.stop()

    with qemu_nbd.open(
            dst_top, "qcow2", read_only=True, backing_chain=False) as c:
        log.debug("dst_top extents: %s", list(nbdutil.extents(c)))

    # Test image content - must match.
    qemu_img.compare(
        src_top, dst_top, format1="qcow2", format2="qcow2", strict=False)

    # Compare allocation for qcow2 chain - nice to have.
    if base_fmt == "qcow2":
        qemu_img.compare(
            src_top, dst_top, format1="qcow2", format2="qcow2", strict=True)


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_download_raw(tmpdir, srv, fmt):
    src = str(tmpdir.join("src"))
    with open(src, "wb") as f:
        f.truncate(IMAGE_SIZE)
        f.seek(IMAGE_SIZE // 2)
        f.write(b"data")

    url = prepare_transfer(srv, "file://" + src)
    dst = str(tmpdir.join("dst"))

    # When we download raw data, we can convert it on-the-fly to other format.
    client.download(url, dst, srv.config.tls.ca_file, fmt=fmt)

    # file backend does not support extents, so downloaded data is always
    # fully allocated.
    qemu_img.compare(src, dst, format1="raw", format2=fmt)


def test_download_qcow2_as_raw(tmpdir, srv):
    src = str(tmpdir.join("src.qcow2"))
    qemu_img.create(src, "qcow2", size=IMAGE_SIZE)

    # Allocate one cluster in the middle of the image.
    with qemu_nbd.open(src, "qcow2") as c:
        c.write(CLUSTER_SIZE, b"a" * CLUSTER_SIZE)
        c.flush()

    actual_size = os.path.getsize(src)
    url = prepare_transfer(srv, "file://" + src, size=actual_size)
    dst = str(tmpdir.join("dst.qcow2"))

    # When downloading qcow2 image using the nbd backend, we get raw data and
    # we can convert it to any format we want. Howver when downloading using
    # the file backend, we get qcow2 bytestream and we cannot convert it.
    #
    # To store the qcow2 bytestream, we must use fmt="raw". This instructs
    # qemu-nbd on the client side to treat the data as raw bytes, storing them
    # without any change on the local file.
    #
    # This is baisically like:
    #
    #   qemu-img convert -f raw -O raw src.qcow2 dst.qcow2
    #
    client.download(url, dst, srv.config.tls.ca_file, fmt="raw")

    # The result should be identical qcow2 image content. Allocation may
    # differ but for this test we get identical allocation.
    qemu_img.compare(src, dst, format1="qcow2", format2="qcow2", strict=True)


@pytest.mark.parametrize("base_fmt", ["raw", "qcow2"])
def test_download_shallow(srv, nbd_server, tmpdir, base_fmt):
    size = 10 * 1024**2

    # Create source base image with some data in first clusters.
    src_base = str(tmpdir.join("src_base." + base_fmt))
    qemu_img.create(src_base, base_fmt, size=size)
    with qemu_nbd.open(src_base, base_fmt) as c:
        c.write(0 * CLUSTER_SIZE, b"a" * CLUSTER_SIZE)
        c.write(1 * CLUSTER_SIZE, b"b" * CLUSTER_SIZE)
        c.write(2 * CLUSTER_SIZE, b"c" * CLUSTER_SIZE)
        c.flush()
        log.debug("src_base extents: %s", list(nbdutil.extents(c)))

    # Create source top image with some data in second cluster and zero in the
    # third cluster.
    src_top = str(tmpdir.join("src_top.qcow2"))
    qemu_img.create(
        src_top, "qcow2", backing_file=src_base, backing_format=base_fmt)
    with qemu_nbd.open(src_top, "qcow2") as c:
        c.write(1 * CLUSTER_SIZE, b"B" * CLUSTER_SIZE)
        c.zero(2 * CLUSTER_SIZE, CLUSTER_SIZE)
        c.flush()

    with qemu_nbd.open(
            src_top, "qcow2", read_only=True, backing_chain=False) as c:
        log.debug("src_top extents: %s", list(nbdutil.extents(c)))

    dst_base = str(tmpdir.join("dst_base." + base_fmt))
    dst_top = str(tmpdir.join("dst_top.qcow2"))

    # Download base image.

    nbd_server.image = src_base
    nbd_server.fmt = base_fmt
    nbd_server.start()

    url = prepare_transfer(srv, nbd_server.sock.url(), size=size)
    client.download(
        url,
        dst_base,
        srv.config.tls.ca_file,
        fmt=base_fmt)

    nbd_server.stop()

    with qemu_nbd.open(dst_base, base_fmt, read_only=True) as c:
        log.debug("dst_base extents: %s", list(nbdutil.extents(c)))

    # Compare image content - must match.
    qemu_img.compare(
        src_base, dst_base, format1=base_fmt, format2=base_fmt, strict=False)

    # And allocation - nice to have.
    if base_fmt == "qcow2" or qemu_nbd.version() >= (6, 0, 0):
        qemu_img.compare(
            src_base, dst_base, format1=base_fmt, format2=base_fmt,
            strict=True)

    # Download top image.

    nbd_server.image = src_top
    nbd_server.fmt = "qcow2"
    nbd_server.backing_chain = False
    nbd_server.start()

    url = prepare_transfer(srv, nbd_server.sock.url(), size=size)
    client.download(
        url,
        dst_top,
        srv.config.tls.ca_file,
        backing_file=dst_base,
        backing_format=base_fmt)

    nbd_server.stop()

    with qemu_nbd.open(
            dst_top, "qcow2", read_only=True, backing_chain=False) as c:
        log.debug("dst_top extents: %s", list(nbdutil.extents(c)))

    # Compare both image content - must match.
    qemu_img.compare(
        src_top, dst_top, format1="qcow2", format2="qcow2", strict=False)

    # And allocation - nice to have.
    qemu_img.compare(
        src_top, dst_top, format1="qcow2", format2="qcow2", strict=True)


def test_upload_proxy_url(tmpdir, srv):
    src = str(tmpdir.join("src"))
    with open(src, "wb") as f:
        f.truncate(IMAGE_SIZE)

    dst = str(tmpdir.join("dst"))
    with open(dst, "wb") as f:
        f.truncate(IMAGE_SIZE)

    # If transfer_url is not accessible, proxy_url is used.
    transfer_url = "https://no.server:54322/images/no-ticket"
    proxy_url = prepare_transfer(srv, "file://" + dst)

    client.upload(src, transfer_url, srv.config.tls.ca_file,
                  proxy_url=proxy_url)

    qemu_img.compare(src, dst, format1="raw", format2="raw", strict=True)


def test_upload_proxy_url_unused(tmpdir, srv):
    src = str(tmpdir.join("src"))
    with open(src, "wb") as f:
        f.truncate(IMAGE_SIZE)

    dst = str(tmpdir.join("dst"))
    with open(dst, "wb") as f:
        f.truncate(IMAGE_SIZE)

    # If transfer_url is accessible, proxy_url is not used.
    transfer_url = prepare_transfer(srv, "file://" + dst)
    proxy_url = "https://no.proxy:54322/images/no-ticket"

    client.upload(src, transfer_url, srv.config.tls.ca_file,
                  proxy_url=proxy_url)

    qemu_img.compare(src, dst, format1="raw", format2="raw", strict=True)


def test_download_proxy_url(tmpdir, srv):
    src = str(tmpdir.join("src"))
    with open(src, "wb") as f:
        f.truncate(IMAGE_SIZE)

    dst = str(tmpdir.join("dst"))

    # If transfer_url is not accessible, proxy_url is used.
    transfer_url = "https://no.server:54322/images/no-ticket"
    proxy_url = prepare_transfer(srv, "file://" + src)

    client.download(transfer_url, dst, srv.config.tls.ca_file, fmt="raw",
                    proxy_url=proxy_url)

    qemu_img.compare(src, dst, format1="raw", format2="raw")


def test_download_proxy_url_unused(tmpdir, srv):
    src = str(tmpdir.join("src"))
    with open(src, "wb") as f:
        f.truncate(IMAGE_SIZE)

    dst = str(tmpdir.join("dst"))

    # If transfer_url is accessible, proxy_url is not used.
    transfer_url = prepare_transfer(srv, "file://" + src)
    proxy_url = "https://no.proxy:54322/images/no-ticket"

    client.download(transfer_url, dst, srv.config.tls.ca_file, fmt="raw",
                    proxy_url=proxy_url)

    qemu_img.compare(src, dst, format1="raw", format2="raw")


def test_progress(tmpdir, srv):
    src = str(tmpdir.join("src"))
    with open(src, "wb") as f:
        f.write(b"b" * 4096)
        f.seek(IMAGE_SIZE // 2)
        f.write(b"b" * 4096)
        f.truncate(IMAGE_SIZE)

    dst = str(tmpdir.join("dst"))
    with open(dst, "wb") as f:
        f.truncate(IMAGE_SIZE)

    url = prepare_transfer(srv, "file://" + dst, sparse=True)

    progress = FakeProgress()
    client.upload(
        src, url, srv.config.tls.ca_file, progress=progress)

    assert progress.size == IMAGE_SIZE

    # Note: when using multiple connections order of updates is not
    # predictable.
    assert set(progress.updates) == {
        # First write.
        4096,
        # First zero.
        IMAGE_SIZE // 2 - 4096,
        # Second write.
        4096,
        # Second zero
        IMAGE_SIZE // 2 - 4096,
    }


def test_progress_callback(tmpdir, srv):
    src = str(tmpdir.join("src"))
    with open(src, "wb") as f:
        f.truncate(IMAGE_SIZE)

    dst = str(tmpdir.join("dst"))
    with open(dst, "wb") as f:
        f.truncate(IMAGE_SIZE)

    url = prepare_transfer(srv, "file://" + dst, size=IMAGE_SIZE, sparse=True)

    progress = []
    client.upload(
        src,
        url,
        srv.config.tls.ca_file,
        progress=progress.append)

    assert progress == [IMAGE_SIZE]


@pytest.mark.parametrize("fmt, compressed", [
    ("raw", False),
    ("qcow2", False),
    ("qcow2", True),
])
def test_info(tmpdir, fmt, compressed):
    # Created temporary file with some data.
    size = 2 * 1024**2
    tmp = str(tmpdir.join("tmp"))
    with open(tmp, "wb") as f:
        f.truncate(size)
        f.write(b"x" * CLUSTER_SIZE)

    # Created test image from temporary file.
    img = str(tmpdir.join("img"))
    qemu_img.convert(tmp, img, "raw", fmt, compressed=compressed)
    img_info = client.info(img)

    # Check image info.
    assert img_info["format"] == fmt
    assert img_info["virtual-size"] == size

    # We don't add member info if member was not specified.
    assert "member-offset" not in img_info
    assert "member-size" not in img_info

    # Create ova with test image.
    member = os.path.basename(img)
    ova = str(tmpdir.join("ova"))
    with tarfile.open(ova, "w") as tar:
        tar.add(img, arcname=member)

    # Get info for the member from the ova.
    ova_info = client.info(ova, member=member)

    # Image info from ova should be the same.
    assert ova_info["format"] == fmt
    assert ova_info["virtual-size"] == size

    # If member was specified, we report also the offset and size.
    with tarfile.open(ova) as tar:
        member_info = tar.getmember(member)
    assert ova_info["member-offset"] == member_info.offset_data
    assert ova_info["member-size"] == member_info.size


@pytest.mark.parametrize("fmt, compressed", [
    ("raw", False),
    ("qcow2", False),
    ("qcow2", True),
])
def test_measure_to_raw(tmpdir, fmt, compressed):
    # Create temporary file with some data.
    size = 2 * 1024**2
    tmp = str(tmpdir.join("tmp"))
    with open(tmp, "wb") as f:
        f.truncate(size)
        f.write(b"x" * CLUSTER_SIZE)

    # Created test image from temporary file.
    img = str(tmpdir.join("img"))
    qemu_img.convert(tmp, img, "raw", fmt, compressed=compressed)

    measure = client.measure(img, "raw")
    assert measure["required"] == size


@pytest.mark.parametrize("fmt, compressed", [
    ("raw", False),
    ("qcow2", False),
    ("qcow2", True),
])
def test_measure_to_qcow2(tmpdir, fmt, compressed):
    # Create temporary file with some data.
    size = 2 * 1024**2
    tmp = str(tmpdir.join("tmp"))
    with open(tmp, "wb") as f:
        f.truncate(size)
        f.write(b"x" * CLUSTER_SIZE)

    # Created test image from temporary file.
    img = str(tmpdir.join("img"))
    qemu_img.convert(tmp, img, "raw", fmt, compressed=compressed)

    measure = client.measure(img, "qcow2")
    assert measure["required"] == 393216


@pytest.mark.parametrize("compressed", [False, True])
@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_measure_from_ova(tmpdir, compressed, fmt):
    # Create temporary file with some data.
    size = 2 * 1024**2
    tmp = str(tmpdir.join("tmp"))
    with open(tmp, "wb") as f:
        f.truncate(size)
        f.write(b"x" * CLUSTER_SIZE)

    # Created test image from temporary file.
    img = str(tmpdir.join("img"))
    qemu_img.convert(tmp, img, "raw", "qcow2", compressed=compressed)

    # Measure the image.
    img_measure = client.measure(img, fmt)

    # We don't add member info if member was not specified.
    assert "member-offset" not in img_measure
    assert "member-size" not in img_measure

    # Add test image to ova.
    member = os.path.basename(img)
    ova = str(tmpdir.join("ova"))
    with tarfile.open(ova, "w") as tar:
        tar.add(img, arcname=member)

    # Measure the image from the ova.
    ova_measure = client.measure(ova, fmt, member=member)

    # Measurement from ova should be same.
    assert ova_measure["required"] == img_measure["required"]
    assert ova_measure["fully-allocated"] == img_measure["fully-allocated"]

    # If member was specified, we report also the offset and size.
    with tarfile.open(ova) as tar:
        member_info = tar.getmember(member)
    assert ova_measure["member-offset"] == member_info.offset_data
    assert ova_measure["member-size"] == member_info.size


@pytest.mark.parametrize("fmt, compressed", [
    ("raw", False),
    ("qcow2", False),
    ("qcow2", True),
])
def test_checksum(tmpdir, fmt, compressed):
    # Create temporary file with some data.
    size = 2 * 1024**2
    tmp = str(tmpdir.join("tmp"))
    with open(tmp, "wb") as f:
        f.truncate(size)
        f.write(b"x" * CLUSTER_SIZE)

    # Create test image from temporary file.
    img = str(tmpdir.join("img"))
    qemu_img.convert(tmp, img, "raw", fmt, compressed=compressed)

    expected = blkhash.checksum(tmp, block_size=1024**2)
    actual = client.checksum(img, block_size=1024**2)
    assert actual == expected


@pytest.mark.parametrize("fmt, compressed", [
    ("raw", False),
    ("qcow2", False),
    ("qcow2", True),
])
def test_checksum_from_ova(tmpdir, fmt, compressed):
    # Create temporary file with some data.
    size = 2 * 1024**2
    tmp = str(tmpdir.join("tmp"))
    with open(tmp, "wb") as f:
        f.truncate(size)
        f.write(b"x" * CLUSTER_SIZE)

    # Create test image from temporary file.
    img = str(tmpdir.join("img"))
    qemu_img.convert(tmp, img, "raw", fmt, compressed=compressed)

    # Add test image to ova.
    member = os.path.basename(img)
    ova = str(tmpdir.join("ova"))
    with tarfile.open(ova, "w") as tar:
        tar.add(img, arcname=member)

    expected = blkhash.checksum(tmp, block_size=1024**2)
    actual = client.checksum(ova, member=member, block_size=1024**2)
    assert actual == expected


@pytest.mark.parametrize("algorithm,digest_size", [
    ("blake2b", 32),
    ("sha1", None),
])
def test_checksum_algorithm(tmpdir, algorithm, digest_size):
    img = str(tmpdir.join("img"))
    qemu_img.create(img, "raw", size="2m")

    expected = blkhash.checksum(
        img, block_size=1024**2, algorithm=algorithm, digest_size=digest_size)
    actual = client.checksum(img, block_size=1024**2, algorithm=algorithm)
    assert actual == expected


def test_zero_extents_raw(tmpdir):
    size = 10 * 1024**2

    # Create image with some data, zero and holes.
    image = str(tmpdir.join("image.raw"))
    qemu_img.create(image, "raw", size=size)
    with qemu_nbd.open(image, "raw") as c:
        c.write(0 * CLUSTER_SIZE, b"A" * CLUSTER_SIZE)
        c.zero(1 * CLUSTER_SIZE, CLUSTER_SIZE)
        c.write(2 * CLUSTER_SIZE, b"B" * CLUSTER_SIZE)
        c.flush()

    extents = list(client.extents(image))

    # Unallocated area in raw image is not reported as a hole.
    assert extents == [
        ZeroExtent(
            start=0 * CLUSTER_SIZE,
            length=CLUSTER_SIZE,
            zero=False,
            hole=False),
        ZeroExtent(
            start=1 * CLUSTER_SIZE,
            length=CLUSTER_SIZE,
            zero=True,
            hole=False),
        ZeroExtent(
            start=2 * CLUSTER_SIZE,
            length=CLUSTER_SIZE,
            zero=False,
            hole=False),
        ZeroExtent(
            start=3 * CLUSTER_SIZE,
            length=size - 3 * CLUSTER_SIZE,
            zero=True,
            hole=False),
    ]


def test_zero_extents_qcow2(tmpdir):
    size = 10 * 1024**2

    # Create base image with one data and one zero cluster.
    base = str(tmpdir.join("base.qcow2"))
    qemu_img.create(base, "qcow2", size=size)
    with qemu_nbd.open(base, "qcow2") as c:
        c.write(0 * CLUSTER_SIZE, b"A" * CLUSTER_SIZE)
        c.zero(1 * CLUSTER_SIZE, CLUSTER_SIZE)
        c.flush()

    # Create top image with one data and one zero cluster.
    top = str(tmpdir.join("top.qcow2"))
    qemu_img.create(
        top, "qcow2", backing_file=base, backing_format="qcow2")
    with qemu_nbd.open(top, "qcow2") as c:
        c.write(3 * CLUSTER_SIZE, b"B" * CLUSTER_SIZE)
        c.zero(4 * CLUSTER_SIZE, CLUSTER_SIZE)
        c.flush()

    extents = list(client.extents(top))

    assert extents == [
        # Extents from base...
        ZeroExtent(
            start=0 * CLUSTER_SIZE,
            length=CLUSTER_SIZE,
            zero=False,
            hole=False),
        ZeroExtent(
            start=1 * CLUSTER_SIZE,
            length=CLUSTER_SIZE,
            zero=True,
            hole=False),
        ZeroExtent(
            start=2 * CLUSTER_SIZE,
            length=CLUSTER_SIZE,
            zero=True,
            hole=True),

        # Extents from top...
        ZeroExtent(
            start=3 * CLUSTER_SIZE,
            length=CLUSTER_SIZE,
            zero=False,
            hole=False),
        ZeroExtent(
            start=4 * CLUSTER_SIZE,
            length=CLUSTER_SIZE,
            zero=True,
            hole=False),

        # Rest of unallocated data...
        ZeroExtent(
            start=5 * CLUSTER_SIZE,
            length=size - 5 * CLUSTER_SIZE,
            zero=True,
            hole=True),
    ]


def test_zero_extents_from_ova(tmpdir):
    size = 10 * 1024**2

    # Create image with data, zero and hole clusters.
    disk = str(tmpdir.join("disk.qcow2"))
    qemu_img.create(disk, "qcow2", size=size)
    with qemu_nbd.open(disk, "qcow2") as c:
        c.write(0 * CLUSTER_SIZE, b"A" * CLUSTER_SIZE)
        c.zero(1 * CLUSTER_SIZE, CLUSTER_SIZE)
        c.flush()

    # Create OVA whith this image.
    ova = str(tmpdir.join("vm.ova"))
    with tarfile.open(ova, "w") as tar:
        tar.add(disk, arcname=os.path.basename(disk))

    extents = list(client.extents(ova, member="disk.qcow2"))

    assert extents == [
        ZeroExtent(
            start=0 * CLUSTER_SIZE,
            length=CLUSTER_SIZE,
            zero=False,
            hole=False),

        # Broken since qemu-nbd 6.0.0.
        ZeroExtent(
            start=1 * CLUSTER_SIZE,
            length=CLUSTER_SIZE,
            zero=True,
            hole=False),

        ZeroExtent(
            start=2 * CLUSTER_SIZE,
            length=size - 2 * CLUSTER_SIZE,
            zero=True,
            hole=True),
    ]


def test_dirty_extents(tmpdir):
    size = 1024**2

    # Create base image with empty dirty bitmap.
    base = str(tmpdir.join("base.qcow2"))
    qemu_img.create(base, "qcow2", size=size)
    qemu_img.bitmap_add(base, "b0")

    # Write data, modifying the dirty bitmap.
    with qemu_nbd.open(base, "qcow2") as c:
        c.write(0 * CLUSTER_SIZE, b"A" * CLUSTER_SIZE)
        c.zero(1 * CLUSTER_SIZE, CLUSTER_SIZE)
        c.flush()

    # Create top image with empty dirty bitmap.
    top = str(tmpdir.join("top.qcow2"))
    qemu_img.create(top, "qcow2", backing_file=base, backing_format="qcow2")
    qemu_img.bitmap_add(top, "b0")

    # Write data, modifying the dirty bitmap.
    with qemu_nbd.open(top, "qcow2") as c:
        c.write(3 * CLUSTER_SIZE, b"B" * CLUSTER_SIZE)
        c.zero(4 * CLUSTER_SIZE, CLUSTER_SIZE)
        c.flush()

    dirty_extents = list(client.extents(base, bitmap="b0"))

    expected = [
        # First cluster is dirty data.
        DirtyExtent(
            start=0 * CLUSTER_SIZE,
            length=1 * CLUSTER_SIZE,
            dirty=True,
            zero=False),
        # Second cluster is dirty zero.
        DirtyExtent(
            start=1 * CLUSTER_SIZE,
            length=1 * CLUSTER_SIZE,
            dirty=True,
            zero=True),
        # Third cluster is clean zero.
        DirtyExtent(
            start=2 * CLUSTER_SIZE,
            length=size - 2 * CLUSTER_SIZE,
            dirty=False,
            zero=True),
    ]

    log.debug("base image dirty extents: %s", dirty_extents)
    assert dirty_extents == expected

    dirty_extents = list(client.extents(top, bitmap="b0"))

    # Note: qemu-nbd reports dirty extents only for the top image, but zero
    # extents are read from the base image.
    expected = [
        # First cluster is clean data, read from base image.
        DirtyExtent(
            start=0 * CLUSTER_SIZE,
            length=1 * CLUSTER_SIZE,
            dirty=False,
            zero=False),
        # Second and third clusters are read from base image. Because they are
        # both clean zero, they are merged.
        DirtyExtent(
            start=1 * CLUSTER_SIZE,
            length=2 * CLUSTER_SIZE,
            dirty=False,
            zero=True),
        # Forth cluster is a data extent modified in top image.
        DirtyExtent(
            start=3 * CLUSTER_SIZE,
            length=1 * CLUSTER_SIZE,
            dirty=True,
            zero=False),
        # Fifth cluster is a zero extent modifed in to image.
        DirtyExtent(
            start=4 * CLUSTER_SIZE,
            length=1 * CLUSTER_SIZE,
            dirty=True,
            zero=True),
        # The rest is clean zero extent.
        DirtyExtent(
            start=5 * CLUSTER_SIZE,
            length=size - 5 * CLUSTER_SIZE,
            dirty=False,
            zero=True),
    ]

    log.debug("top image dirty extents: %s", dirty_extents)
    assert dirty_extents == expected


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_stress(srv, nbd_server, tmpdir, fmt):
    size = 10 * 1024**2

    # Create empty source and destination images.
    src = str(tmpdir.join("src." + fmt))
    qemu_img.create(src, fmt, size=size)
    dst = str(tmpdir.join("dst." + fmt))
    qemu_img.create(dst, fmt, size=size)

    # Upload and download the image multiple times. This used to fail randomly
    # when the executor closed the destination backend before it was cloned by
    # the workers.

    nbd_server.image = dst
    nbd_server.fmt = fmt
    nbd_server.start()

    url = prepare_transfer(srv, nbd_server.sock.url(), size=size)

    for i in range(20):
        client.upload(src, url, srv.config.tls.ca_file)
        client.download(url, src, srv.config.tls.ca_file, fmt=fmt)

    nbd_server.stop()


def test_concurrent_downloads(srv, tmpdir):
    # Testing that we can serve 10 conccurent transfers, assuming 4
    # connections per client.
    # https://bugzilla.redhat.com/2066113

    size = 10 * 1024**2

    def download(url, dst):
        client.download(url, dst, srv.config.tls.ca_file)

    downloads = []
    try:
        for i in range(10):
            src = str(tmpdir.join(f"{i:02d}.src.qcow2"))
            qemu_img.create(src, "qcow2", size=size)

            dst = str(tmpdir.join(f"{i:02d}.dst.qcow2"))
            qemu_img.create(dst, "qcow2", size=size)

            sock = str(tmpdir.join(f"{i:02d}.sock"))

            nbd_server = qemu_nbd.Server(src, "qcow2", nbd.UnixAddress(sock))
            url = prepare_transfer(srv, nbd_server.sock.url(), size=size)
            nbd_server.start()

            downloads.append((nbd_server, url, dst))

        with ThreadPoolExecutor(max_workers=10) as executor:
            tasks = [executor.submit(download, url, dst)
                     for _, url, dst in downloads]
            for t in as_completed(tasks):
                t.result()
    finally:
        for nbd_server, _, _ in downloads:
            nbd_server.stop()



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import time
import pytest

from urllib.parse import urlparse

from ovirt_imageio._internal import extent
from ovirt_imageio._internal import qemu_img
from ovirt_imageio._internal import qemu_nbd
from ovirt_imageio._internal.backends import nbd, memory
from ovirt_imageio._internal.nbd import UnixAddress

from ovirt_imageio.client import _io

ZERO_PARAMS = [
    # Copying to image with unknown content.
    pytest.param(True, True, id="unknown-image"),

    # Copying to new empty image.
    pytest.param(False, True, id="new-image"),

    # Copying to new empty image with a backing file.
    pytest.param(True, False, id="backing-file"),
]

CHUNK_SIZE = 512


class FakeProgress:

    def __init__(self):
        self.updates = []

    def update(self, n):
        self.updates.append(n)


@pytest.mark.parametrize("src_fmt,dst_fmt", [
    ("raw", "raw"),
    ("raw", "qcow2"),
    ("qcow2", "qcow2"),
    ("qcow2", "raw"),
])
@pytest.mark.parametrize("zero,hole", ZERO_PARAMS)
def test_copy_nbd_to_nbd(tmpdir, src_fmt, dst_fmt, zero, hole):
    # Default cluser size with qcow2 format.
    cluster_size = 64 * 1024
    extents = [
        ("data", cluster_size),
        ("zero", cluster_size),
        ("data", cluster_size),
        ("hole", cluster_size + _io.MAX_ZERO_SIZE),
        ("data", cluster_size + _io.BUFFER_SIZE),
        ("hole", cluster_size),
        ("data", cluster_size),
    ]
    size = sum(length for _, length in extents)

    src = str(tmpdir.join("src." + src_fmt))
    qemu_img.create(src, src_fmt, size=size)
    populate_image(src, src_fmt, extents)

    src_sock = UnixAddress(tmpdir.join("src.sock"))
    src_url = urlparse(src_sock.url())

    dst = str(tmpdir.join("dst." + dst_fmt))
    qemu_img.create(dst, dst_fmt, size=size)
    dst_sock = UnixAddress(tmpdir.join("dst.sock"))
    dst_url = urlparse(dst_sock.url())

    # Note: We need extra worker for reading extents for source.
    max_workers = 2
    with qemu_nbd.run(
                src, src_fmt, src_sock,
                read_only=True,
                shared=max_workers + 1), \
            qemu_nbd.run(
                dst, dst_fmt, dst_sock,
                shared=max_workers), \
            nbd.open(src_url, "r") as src_backend, \
            nbd.open(dst_url, "r+", sparse=True) as dst_backend:

        # Because we copy to new image, we can always use zero=False, but we
        # test both to verify that the result is the same.
        _io.copy(
            src_backend,
            dst_backend,
            max_workers=max_workers,
            zero=zero,
            hole=hole)

    # Compare image content - must match.
    qemu_img.compare(src, dst)

    # Allocation can be compared only with qcow2 images when we write zeroes to
    # zero extents and skip holes.
    if src_fmt == "qcow2" and dst_fmt == "qcow2" and zero and not hole:
        qemu_img.compare(src, dst, strict=True)


@pytest.mark.parametrize("buffer_size", [128, 1024])
@pytest.mark.parametrize("zero,hole", ZERO_PARAMS)
@pytest.mark.parametrize("progress", [None, FakeProgress()])
def test_copy_generic(buffer_size, zero, hole, progress):
    src_extents = create_zero_extents("B0-")
    src_backing = create_backing("B0-")

    dst_backing = create_backing(
        "AAA" if zero and hole else "AA0" if zero else "A00")

    src = memory.Backend(
        mode="r", data=src_backing, extents={"zero": src_extents})

    dst = memory.Backend("r+", data=dst_backing)

    _io.copy(
        src, dst,
        max_workers=1,
        buffer_size=buffer_size,
        zero=zero,
        hole=hole,
        progress=progress)

    assert dst_backing == src_backing


@pytest.mark.parametrize("buffer_size", [128, 1024])
@pytest.mark.parametrize("zero,hole", ZERO_PARAMS)
@pytest.mark.parametrize("progress", [None, FakeProgress()])
def test_copy_read_from(buffer_size, zero, hole, progress):
    src_extents = create_zero_extents("B0-")
    src_backing = create_backing("B0-")

    dst_backing = create_backing(
        "AAA" if zero and hole else "AA0" if zero else "A00")

    src = memory.Backend(
        mode="r", data=src_backing, extents={"zero": src_extents})

    dst = memory.ReaderFrom("r+", data=dst_backing)

    _io.copy(
        src, dst,
        max_workers=1,
        buffer_size=buffer_size,
        zero=zero,
        hole=hole,
        progress=progress)

    assert dst_backing == src_backing


@pytest.mark.parametrize("buffer_size", [128, 1024])
@pytest.mark.parametrize("zero,hole", ZERO_PARAMS)
@pytest.mark.parametrize("progress", [None, FakeProgress()])
def test_copy_write_to(buffer_size, zero, hole, progress):
    src_extents = create_zero_extents("B0-")
    src_backing = create_backing("B0-")

    dst_backing = create_backing(
        "AAA" if zero and hole else "AA0" if zero else "A00")

    src = memory.WriterTo(
        mode="r", data=src_backing, extents={"zero": src_extents})

    dst = memory.Backend("r+", data=dst_backing)

    _io.copy(
        src, dst,
        max_workers=1,
        buffer_size=buffer_size,
        zero=zero,
        hole=hole,
        progress=progress)

    assert dst_backing == src_backing


@pytest.mark.parametrize("progress", [None, FakeProgress()])
def test_copy_dirty(progress):
    src = memory.Backend(
        mode="r",
        data=create_backing("ABCD"),
        extents={"dirty": create_dirty_extents("AbCd")},
    )
    dst_backing = create_backing("0000")
    dst = memory.Backend("r+", data=dst_backing)

    _io.copy(src, dst, dirty=True, max_workers=1, progress=progress)

    # Copy dirty extents, skip clean extents.
    assert dst_backing == create_backing("A0C0")


@pytest.mark.parametrize("zero,hole", ZERO_PARAMS)
def test_copy_data_progress(zero, hole):
    src = memory.Backend(
        mode="r",
        data=create_backing("A0C-"),
        extents={"zero": create_zero_extents("A0C-")},
    )
    dst_backing = create_backing("0000")
    dst = memory.Backend("r+", data=dst_backing)

    p = FakeProgress()
    _io.copy(src, dst, max_workers=1, zero=zero, hole=hole, progress=p)

    # Report at least every extent.
    assert len(p.updates) >= 4

    # Report entire image size.
    assert sum(p.updates) == len(dst_backing)


def test_copy_dirty_progress():
    src = memory.Backend(
        mode="r",
        data=create_backing("A0C-"),
        extents={"dirty": create_dirty_extents("A0C-")},
    )
    dst_backing = create_backing("0000")
    dst = memory.Backend("r+", data=dst_backing)

    p = FakeProgress()
    _io.copy(src, dst, dirty=True, max_workers=1, progress=p)

    # Report at least every extent.
    assert len(p.updates) >= 4

    # Report entire image size.
    assert sum(p.updates) == len(dst_backing)


class BackendError(Exception):
    pass


class FailingBackend:

    def __init__(self, fail_read=False, fail_write=False, delay=0.1):
        self.fail_read = fail_read
        self.fail_write = fail_write
        self.delay = delay

    def clone(self):
        return FailingBackend(
            fail_read=self.fail_read, fail_write=self.fail_write)

    def size(self):
        return 10 * 1024**3

    def extents(self, ctx="zero"):
        return [extent.ZeroExtent(0, self.size(), False, False)]

    def readinto(self, buf):
        time.sleep(self.delay)
        if self.fail_read:
            raise BackendError("read error")
        return len(buf)

    def write(self, buf):
        time.sleep(self.delay)
        if self.fail_write:
            raise BackendError("write error")
        return len(buf)

    def seek(self, n, how=None):
        pass

    def close(self):
        pass


def test_reraise_dst_error():
    src = FailingBackend()
    dst = FailingBackend(fail_write=True)
    with pytest.raises(BackendError) as e:
        _io.copy(src, dst)
    assert str(e.value) == "write error"


def test_reraise_src_error():
    src = FailingBackend(fail_read=True)
    dst = FailingBackend()
    with pytest.raises(BackendError) as e:
        _io.copy(src, dst)
    assert str(e.value) == "read error"


# Testing helpers.

def populate_image(path, fmt, extents):
    with qemu_nbd.open(path, fmt) as c:
        offset = 0
        for kind, length in extents:
            if kind == "data":
                c.write(offset, b"x" * length)
            elif kind == "zero":
                c.zero(offset, length, punch_hole=False)
            elif kind == "hole":
                pass  # Unallocated
            offset += length
        c.flush()


def create_zero_extents(fmt):
    """
    Create zero extents from format string.

    "A0-" -> [
        ZeroExtent(0 * CHUNK_SIZE, CHUNK_SIZE, False, False),
        ZeroExtent(1 * CHUNK_SIZE, CHUNK_SIZE, True, False),
        ZeroExtent(2 * CHUNK_SIZE, CHUNK_SIZE, True, True),
    ]
    """
    extents = []
    offset = 0

    for c in fmt:
        if c == "0":
            extents.append(
                extent.ZeroExtent(offset, CHUNK_SIZE, zero=True, hole=False))
        elif c == "-":
            extents.append(
                extent.ZeroExtent(offset, CHUNK_SIZE, zero=True, hole=True))
        else:
            extents.append(
                extent.ZeroExtent(offset, CHUNK_SIZE, zero=False, hole=False))
        offset += CHUNK_SIZE

    return extents


def create_dirty_extents(fmt):
    """
    Create dirty extents from format string.

    "Ab" -> [
        DirtyExtent(0 * CHUNK_SIZE, CHUNK_SIZE, True, False),
        DirtyExtent(1 * CHUNK_SIZE, CHUNK_SIZE, False, False),
    ]
    """
    extents = []
    offset = 0

    for c in fmt:
        extents.append(
            extent.DirtyExtent(
                offset, CHUNK_SIZE, dirty=c.isupper(), zero=False))
        offset += CHUNK_SIZE

    return extents


def create_backing(fmt):
    """
    Create backing from format string.

    "A0-" -> bytearray(
        b"A" * CHUNK_SIZE + b"\0" * CHUNK_SIZE + b"\0" * CHUNK_SIZE
    )
    """
    b = bytearray()

    for c in fmt:
        if c in ("0", "-"):
            b += b"\0" * CHUNK_SIZE
        else:
            b += c.encode("ascii") * CHUNK_SIZE

    return b



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

from urllib.parse import urlunparse
import pytest

from ovirt_imageio._internal import auth
from ovirt_imageio._internal import backends
from ovirt_imageio._internal import config
from ovirt_imageio._internal import errors
from ovirt_imageio._internal import nbd

from . import testutil
from . marks import flaky_in_ovirt_ci


class Request:

    connection_id = 1

    def __init__(self):
        self.context = {}
        self.connection_timeout = None

    def set_connection_timeout(self, timeout):
        self.connection_timeout = timeout


@pytest.fixture
def cfg():
    return config.load([])


def test_get_caching(tmpurl, cfg):
    ticket = auth.Ticket(
        testutil.create_ticket(url=urlunparse(tmpurl)), cfg)
    req = Request()
    c1 = backends.get(req, ticket, cfg)

    # Context is cached in the ticket.
    assert ticket.get_context(req.connection_id) is c1
    assert c1.backend.name == "file"
    assert len(c1.buffer) == cfg.backend_file.buffer_size

    # Next call return the cached instance.
    c2 = backends.get(req, ticket, cfg)
    assert c1 is c2

    # Closing req.context removes the context from the ticket.
    req.context[ticket.uuid].close()

    c3 = backends.get(req, ticket, cfg)
    assert c2.backend.name == "file"
    assert c3 is not c1


def test_get_set_timeout(tmpurl, cfg):
    ticket = auth.Ticket(
        testutil.create_ticket(
            url=urlunparse(tmpurl),
            inactivity_timeout=300),
        cfg)

    req = Request()
    assert req.connection_timeout is None

    # Authorizing the ticket set connection timeout.
    backends.get(req, ticket, cfg)
    assert req.connection_timeout == ticket.inactivity_timeout


def test_get_canceled_ticket(tmpurl, cfg):
    ticket = auth.Ticket(
        testutil.create_ticket(url=urlunparse(tmpurl)), cfg)
    req = Request()
    ticket.cancel()

    # If the ticket was canceled, getting a backend raises.
    with pytest.raises(errors.AuthorizationError):
        backends.get(req, ticket, cfg)

    # And nothing is stored in the request context.
    assert req.context == {}


@pytest.mark.parametrize("ops,readable,writable", [
    (["read"], True, False),
    (["read", "write"], True, True),
    (["write"], True, True),
])
def test_get_ops(tmpurl, cfg, ops, readable, writable):
    ticket = auth.Ticket(
        testutil.create_ticket(url=urlunparse(tmpurl), ops=ops), cfg)
    req = Request()
    b = backends.get(req, ticket, cfg).backend

    # Create a read-write file backend.
    assert b.name == "file"
    assert b.readable() == readable
    assert b.writable() == writable


@pytest.mark.parametrize("sparse", [True, False])
def test_get_sparse(tmpurl, cfg, sparse):
    ticket = auth.Ticket(
        testutil.create_ticket(url=urlunparse(tmpurl), sparse=sparse), cfg)
    req = Request()
    b = backends.get(req, ticket, cfg).backend

    assert b.name == "file"
    assert b.sparse == sparse


@pytest.mark.parametrize("transport", [
    "unix",
    pytest.param("tcp", marks=flaky_in_ovirt_ci),
])
def test_get_nbd_backend(tmpdir, cfg, nbd_server, transport):
    if transport == "unix":
        nbd_server.sock = nbd.UnixAddress(tmpdir.join("sock"))
    else:
        nbd_server.sock = nbd.TCPAddress(
            "localhost", testutil.random_tcp_port())
    nbd_server.start()

    ticket = auth.Ticket(
        testutil.create_ticket(url=urlunparse(nbd_server.url)), cfg)
    req = Request()
    b = backends.get(req, ticket, cfg).backend

    assert b.name == "nbd"



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
To run the demo server:

$ python test/httpdemo.py

Using /echo endpoint:

$ curl -X PUT --upload-file fedora-27.img \
    http://localhost:8000/echo/ticket > /dev/null
  % Total    % Received % Xferd  Average Speed   Time    Time     Time  Current
                                 Dload  Upload   Total   Spent    Left  Speed
100 12.0G  100 6144M  100 6144M   659M   659M  0:00:09  0:00:09 --:--:-- 1313M

Benchmarking the server with wrk[1]:

$ ./wrk -t1 -c4 --latency http://localhost:8000/bench/4
Running 10s test @ http://localhost:8000/bench/4
  1 threads and 4 connections
  Thread Stats   Avg      Stdev     Max   +/- Stdev
    Latency   763.47us  257.50us   5.16ms   74.47%
    Req/Sec     5.27k    95.89     5.39k    90.00%
  Latency Distribution
     50%  719.00us
     75%    0.89ms
     90%    1.09ms
     99%    1.57ms
  52410 requests in 10.00s, 5.65MB read
Requests/sec:   5240.66
Transfer/sec:    578.34KB

[1] https://github.com/wg/wrk

"""

import argparse
import logging
from ovirt_imageio._internal import http
from ovirt_imageio._internal import stats

log = logging.getLogger("httpdemo")


class Echo:

    def put(self, req, resp, ticket):
        if req.headers.get("expect") == "100-continue":
            resp.send_info(http.CONTINUE)

        count = req.content_length
        resp.headers["content-length"] = count

        while count:
            with req.clock.run("read"):
                chunk = req.read(1024 * 1024)
            if not chunk:
                raise http.Error(400, "Client disconnected")
            with req.clock.run("write"):
                resp.write(chunk)
            count -= len(chunk)


class Bench:

    def get(self, req, resp, name):
        body = b"%s\n" % name.encode("utf-8")
        resp.headers["content-length"] = len(body)
        with req.clock.run("write"):
            resp.write(body)


class Stream:

    def get(self, req, resp, count):
        count = int(count) * 1024**2
        resp.headers["content-length"] = count
        with open("/dev/zero", "rb") as f:
            while count:
                with req.clock.run("read"):
                    chunk = f.read(min(count, 1024**2))
                with req.clock.run("write"):
                    resp.write(chunk)
                count -= len(chunk)

    def put(self, req, resp, name):
        count = req.content_length
        with open("/dev/null", "wb") as f:
            while count:
                with req.clock.run("read"):
                    chunk = req.read(1024 * 1024)
                if not chunk:
                    raise http.Error(400, "Client disconnected")
                with req.clock.run("write"):
                    f.write(chunk)
                count -= len(chunk)


parser = argparse.ArgumentParser()
parser.add_argument(
    "-d", "--debug",
    action="store_true",
    help="Debug mode")
parser.add_argument(
    "-p", "--port",
    default=8000,
    help="listen port (default 8000)")
args = parser.parse_args()

logging.basicConfig(
    level=logging.DEBUG if args.debug else logging.INFO,
    format="%(asctime)s %(levelname)-7s (%(threadName)s) %(message)s")

log.info("Starting server on port %s", args.port)

server = http.Server(("", args.port), http.Connection)
if args.debug:
    server.clock_class = stats.Clock

server.app = http.Router([
    (r"/bench/(.*)", Bench()),
    (r"/stream/(.*)", Stream()),
    (r"/echo/(.*)", Echo()),
])

server.serve_forever()



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

class Error(Exception):
    msg = "Overide this in a subclass"

    def __str__(self):
        return self.msg.format(self=self)


class PartialContent(Error):
    msg = "Requested {self.requested} bytes, available {self.available} bytes"

    def __init__(self, requested, available):
        self.requested = requested
        self.available = available


class InvalidTicket(Error):
    """Base class for ticket errors"""


class MissingTicketParameter(InvalidTicket):
    msg = "Required ticket parameter is missing: {self.parameter}"

    def __init__(self, parameter):
        self.parameter = parameter


class InvalidTicketParameter(InvalidTicket):
    msg = ("Invalid value for {self.parameter!r}: "
           "{self.value!r}: {self.reason}")

    def __init__(self, parameter, value, reason):
        self.parameter = parameter
        self.value = value
        self.reason = reason


class AuthorizationError(Error):
    msg = "You are not allowed to access this resource: {self.reason}"

    def __init__(self, reason):
        self.reason = reason


class TransferCancelTimeout(Error):
    msg = "Timeout cancelling transfer {self.transfer_id}"

    def __init__(self, transfer_id):
        self.transfer_id = transfer_id


class UnsupportedOperation(Error):
    msg = "Operation not supported: {self.reason}"

    def __init__(self, reason):
        self.reason = reason


class TlsConfigurationError(Error):
    msg = ("TLS enabled (see [tls] section in daemon.conf), but not "
           "configured: ca_file = {self.cfg.ca_file}, cert_file = "
           "{self.cfg.cert_file}, key_file = {self.cfg.key_file}")

    def __init__(self, cfg):
        self.cfg = cfg


class InvalidConfig(Error):
    msg = "Invalid configuration: {self.key} = {self.value}"

    def __init__(self, key, value):
        self.key = key
        self.value = value


class ServerStartupError(Error):
    msg = "Server failed to start: {self.reason}"

    def __init__(self, reason):
        self.reason = reason



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import json

import pytest

from ovirt_imageio._internal import config
from ovirt_imageio._internal import server
from ovirt_imageio._internal import version

from .. import http


@pytest.fixture(scope="module")
def srv():
    cfg = config.load(["test/conf/daemon.conf"])
    s = server.Server(cfg)
    s.start()
    yield s
    s.stop()


def test_get(srv):
    with http.RemoteClient(srv.config) as c:
        res = c.get("/info/")
        data = res.read()

    assert res.status == 200
    assert json.loads(data) == {"version": version.string}
    assert res.getheader("Access-Control-Allow-Origin") is None


def test_cors(srv):
    headers = {"Origin": "https://foo.example"}
    with http.RemoteClient(srv.config) as c:
        res = c.get("/info/", headers=headers)
        res.read()

    assert res.status == 200
    assert res.getheader("Access-Control-Allow-Origin") == "*"



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import errno
import http.client as http_client
import io
import json
import logging
import os
import socket
import time

from contextlib import closing
from contextlib import contextmanager

import pytest

from ovirt_imageio._internal import http
from ovirt_imageio._internal import util
from ovirt_imageio._internal import version

log = logging.getLogger("test")

ERROR_CONTENT_TYPE = "text/plain; charset=UTF-8"


class Demo:

    def get(self, req, resp, name):
        body = b"%s\n" % name.encode("utf-8")
        resp.headers["content-length"] = len(body)
        # Test handling for UTF-8 values (Hebrew Alef). This usage is not
        # correct according to HTTP RFCs, but it seems to be supported by the
        # browsers we tested.  See https://tools.ietf.org/html/rfc5987 for a
        # more correct way to do this.
        cd = b"attachment; filename=\xd7\x90".decode("utf-8")
        resp.headers["content-disposition"] = cd
        resp.write(body)

    def delete(self, req, resp, name):
        resp.status_code = http.NO_CONTENT

    def options(self, req, resp, name):
        resp.headers["allow"] = "GET,DELETE,OPTIONS"


class EchoRead:

    def put(self, req, resp, ticket):
        if req.headers.get("expect") == "100-continue":
            resp.send_info(http.CONTINUE)

        count = req.content_length
        resp.headers["content-length"] = count

        while count:
            chunk = req.read(1024 * 1024)
            if not chunk:
                raise http.Error(http.BAD_REQUEST, "Client disconnected")
            resp.write(chunk)
            count -= len(chunk)


class EchoReadinto:

    def put(self, req, resp, ticket):
        if req.headers.get("expect") == "100-continue":
            resp.send_info(http.CONTINUE)

        count = req.content_length
        resp.headers["content-length"] = count

        buf = bytearray(1024**2)
        with memoryview(buf) as view:
            while count:
                n = req.readinto(buf)
                if not n:
                    raise http.Error(http.BAD_REQUEST, "Client disconnected")
                resp.write(view[:n])
                count -= n


class JSON:

    def put(self, req, resp):
        msg = json.loads(req.read())
        resp.send_json(msg)


class RangeDemo:
    """
    Demonstrate using Range and Content-Range headers.
    """

    def __init__(self):
        self.file = io.BytesIO()

    def get(self, req, resp):
        complete = self.file.seek(0, os.SEEK_END)
        offset = 0
        size = complete

        # Handle range request.
        if req.range:
            offset = req.range.first
            if offset < 0:
                offset += complete
                size = complete - offset
            elif req.range.last is not None:
                size = req.range.last - offset + 1
            else:
                size = complete - offset

        if offset + size > complete:
            raise http.Error(
                http.REQUESTED_RANGE_NOT_SATISFIABLE,
                "Requested {} bytes, available {} bytes"
                .format(size, complete - offset),
                content_range="bytes */{}".format(complete - offset))

        resp.headers["content-length"] = size

        if req.range:
            resp.status_code = http.PARTIAL_CONTENT
            resp.headers["content-range"] = "bytes %d-%d/%d" % (
                offset, offset + size - 1, complete)

        self.file.seek(offset)
        resp.write(self.file.read(size))

    def put(self, req, resp):
        offset = req.content_range.first if req.content_range else 0
        self.file.seek(offset)
        self.file.write(req.read())
        if req.length != 0:
            raise http.Error(http.BAD_REQUEST, "Unexpected EOF")


class RequestInfo:

    def get(self, req, resp, arg=None):
        self.send_response(req, resp, arg)

    def put(self, req, resp, arg=None):
        body = req.read().decode("utf-8")
        self.send_response(req, resp, arg, body)

    def send_response(self, req, resp, arg, content=None):
        # Python 2.7 returns lowercase keys, Python 3.6 keeps original
        # case. Since headers are case insensitive, lets normalize both to
        # lowercase.
        headers = dict((k.lower(), req.headers[k]) for k in req.headers)

        # Ensure that req.query return decoded values. This is hard to test
        # using json response, since json accepts both text and bytes and
        # generate a bytestream that decodes to unicode on the other side.
        for k, v in req.query.items():
            assert type(k) == str
            assert type(v) == str

        # Simple values
        info = {
            "method": req.method,
            "uri": req.uri,
            "arg": arg,
            "path": req.path,
            "query": req.query,
            "version": req.version,
            "content_length": req.content_length,
            "content": content,
            "headers": headers,
            "client_addr": req.client_addr,
        }

        # Complex values
        if req.range:
            info["range"] = {"first": req.range.first,
                             "last": req.range.last}
        else:
            info["range"] = req.range
        if req.content_range:
            info["content_range"] = {"first": req.content_range.first,
                                     "last": req.content_range.last,
                                     "complete": req.content_range.complete}
        else:
            info["content_range"] = req.content_range

        resp.send_json(info)


class Context:
    """
    Keep per-connection state example.
    """

    def put(self, req, resp, name):
        value = req.read()
        req.context[name] = value

    def get(self, req, resp, name):
        if name not in req.context:
            raise http.Error(http.NOT_FOUND, "No such name {!r}".format(name))
        value = req.context[name]
        resp.headers["content-length"] = len(value)
        resp.write(value)

    def delete(self, req, resp, name):
        req.context.pop(name, None)
        resp.status_code = http.NO_CONTENT


class Closeable:

    def __init__(self, name, log):
        self.name = name
        self.log = log

    def close(self):
        self.log.write("{} was closed\n".format(self.name))
        # For checking that all objects are closed when a connection is closed.
        raise RuntimeError("Error closing {!r}".format(self.name))


class CloseContext:
    """
    Example for closing objects when connection is closed.
    """

    def __init__(self):
        self.log = io.StringIO()

    def put(self, req, resp, name):
        req.context[name] = Closeable(name, self.log)

    def get(self, req, resp, *args):
        value = self.log.getvalue().encode("utf-8")
        self.log = io.StringIO()
        resp.headers["content-length"] = len(value)
        resp.write(value)


class ServerError:

    def get(self, req, resp, name):
        raise RuntimeError("secret data")

    def put(self, req, resp, name):
        # Raising without reading payload wil fail with EPIPE on the
        # client side. If the client is careful, it will get error 500.
        raise RuntimeError("secret data")


class ServerSocketError:

    def get(self, req, resp, name):
        # Fake a fatal socket error that is not related to the HTTP connection.
        # An example error is a backend failing to connect to a remote server.
        # We expect to get INTERNAL_SERVER_ERROR on the client side.
        # Until 1.5.2, this error was wrongly handled as an error on the HTTP
        # connection, and we closed the connection silently instead of
        # returning an error to the client.
        raise socket.error(errno.ECONNRESET, "fake socket error")

    put = get


class ClientError:

    def get(self, req, resp, name):
        raise http.Error(http.FORBIDDEN, "No data for you!")

    def put(self, req, resp, name):
        # Raising without reading payload wil fail with EPIPE on the
        # client side. If the client is careful, it will get error 403.
        raise http.Error(http.FORBIDDEN, "No data for you!")


class KeepConnection:

    def put(self, req, resp):
        # Fail after reading the entire request payload, so the server
        # should keep the connection open.
        req.read()
        raise http.Error(http.FORBIDDEN, "No data for you!")


class PartialResponse:

    def get(self, req, resp):
        # Fail after sending the first part of the response. The
        # connection shold be closed.
        resp.headers["content-length"] = 1000
        resp.write(b"Starting response...")
        raise http.Error(http.INTERNAL_SERVER_ERROR, "No more data for you!")


@contextmanager
def demo_server(address, prefer_ipv4=False):
    server = http.Server(
        (address, 0), http.Connection, prefer_ipv4=prefer_ipv4)
    log.info("Server listening on %r", server.server_address)
    server.app = http.Router([(r"/demo/(.*)", Demo())])

    t = util.start_thread(
        server.serve_forever,
        kwargs={"poll_interval": 0.1})
    try:
        yield server
    finally:
        server.shutdown()
        t.join()


@pytest.fixture(scope="module")
def server():
    server = http.Server(("127.0.0.1", 0), http.Connection)
    log.info("Server listening on %r", server.server_address)

    server.app = http.Router([
        (r"/demo/(.*)", Demo()),
        (r"/echo-read/(.*)", EchoRead()),
        (r"/echo-readinto/(.*)", EchoReadinto()),
        (r"/json/", JSON()),
        (r"/range-demo/", RangeDemo()),
        (r"/request-info/(.*)", RequestInfo()),
        (r"/context/(.*)", Context()),
        (r"/close-context/(.*)", CloseContext()),
        (r"/server-error/(.*)", ServerError()),
        (r"/server-socket-error/(.*)", ServerSocketError()),
        (r"/client-error/(.*)", ClientError()),
        (r"/keep-connection/", KeepConnection()),
        (r"/partial-response/", PartialResponse()),
    ])

    t = util.start_thread(
        server.serve_forever,
        kwargs={"poll_interval": 0.1})
    try:
        yield server
    finally:
        server.shutdown()
        t.join()


def test_demo_headers(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", "/demo/name")
        r = con.getresponse()
        r.read()
        assert r.getheader("Server") == "imageio/" + version.string


def test_demo_get(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", "/demo/name")
        r = con.getresponse()
        assert r.status == http.OK
        assert r.read() == b"name\n"


def test_demo_get_utf8_headers(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", "/demo/name")
        r = con.getresponse()
        r.read()
        assert r.status == http.OK
        cd = r.getheader("content-disposition")
        # HTTP permit only ASCII for headers content, so python 3 decode
        # headers using latin1. This fixes the bad decoding.
        cd = cd.encode("latin1")
        cd = cd.decode("utf-8")
        assert cd == b"attachment; filename=\xd7\x90".decode("utf-8")


def test_demo_max_request_length(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        # GET /demo/xxxxxxxxxx... HTTP/1.1\r\n
        name = "x" * 4075
        con.request("GET", "/demo/" + name)
        r = con.getresponse()
        assert r.status == http.OK
        assert r.read().decode("ascii") == name + "\n"


def test_demo_request_length_too_long(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        # GET /demo/xxxxxxxxxx... HTTP/1.1\r\n
        con.request("GET", "/demo/" + "x" * 4076)
        r = con.getresponse()
        assert r.status == http.REQUEST_URI_TOO_LARGE
        r.read()


def test_demo_get_empty(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", "/demo/")
        r = con.getresponse()
        assert r.status == http.OK
        assert r.read() == b"\n"


def test_demo_delete(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("DELETE", "/demo/name")
        r = con.getresponse()
        assert r.status == http.NO_CONTENT
        assert r.read() == b""


def test_demo_options(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("OPTIONS", "/demo/name")
        r = con.getresponse()
        assert r.status == http.OK
        assert r.getheader("allow") == "GET,DELETE,OPTIONS"
        assert r.read() == b""


@pytest.mark.parametrize("data", [b"it works!", b""])
def test_echo_read(server, data):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("PUT", "/echo-read/test", body=data)
        r = con.getresponse()
        assert r.status == http.OK
        assert r.read() == data


@pytest.mark.parametrize("data", [b"it works!", b""])
def test_echo_readinto(server, data):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("PUT", "/echo-readinto/test", body=data)
        r = con.getresponse()
        assert r.status == http.OK
        assert r.read() == data


def test_echo_100_continue(server):
    data = b"it works!"
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request(
            "PUT",
            "/echo-read/test",
            body=data,
            headers={"expect": "100-continue"})
        r = con.getresponse()
        assert r.status == http.OK
        assert r.read() == data


def test_json(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        send = {"Hebrew Alef": "\u05d0"}
        data = json.dumps(send).encode("utf-8") + b"\n"
        con.request("PUT", "/json/", body=data)
        r = con.getresponse()
        data = r.read()
        assert r.status == http.OK
        assert r.getheader("content-type") == "application/json"
        assert int(r.getheader("content-length")) == len(data)
        assert data.endswith(b"\n")
        recv = json.loads(data)
        assert send == recv


def test_range_demo(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        # Get complete resource, should return 0 bytes.
        con.request("GET", "/range-demo/")
        r = con.getresponse()
        body = r.read()
        assert r.status == http.OK
        assert body == b""

        # Put some data.
        con.request("PUT", "/range-demo/", body=b"it works!")
        r = con.getresponse()
        r.read()
        assert r.status == http.OK

        # Get part of the resource.
        con.request("GET", "/range-demo/", headers={"range": "bytes=3-"})
        r = con.getresponse()
        body = r.read()
        assert r.status == http.PARTIAL_CONTENT
        assert body == b"works!"

        # Replace part of the resource.
        con.request("PUT", "/range-demo/",
                    body=b"really works!",
                    headers={"content-range": "bytes 3-*/*"})
        r = con.getresponse()
        r.read()
        assert r.status == http.OK

        # Get last bytes of the resource.
        con.request("GET", "/range-demo/", headers={"range": "bytes=-13"})
        r = con.getresponse()
        body = r.read()
        assert r.status == http.PARTIAL_CONTENT
        assert body == b"really works!"

        # Get invalid range after the last byte.
        con.request("GET", "/range-demo/", headers={"range": "bytes=0-100"})
        r = con.getresponse()
        r.read()
        assert r.status == http.REQUESTED_RANGE_NOT_SATISFIABLE
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE
        assert r.getheader("content-range") == "bytes */16"


def test_request_info_get(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", "/request-info/arg")
        r = con.getresponse()
        assert r.status == http.OK
        info = json.loads(r.read())

    assert info["method"] == "GET"
    assert info["uri"] == "/request-info/arg"
    assert info["path"] == "/request-info/arg"
    assert info["arg"] == "arg"
    assert info["query"] == {}
    assert info["version"] == "HTTP/1.1"
    assert info["content_length"] is None
    assert info["content"] is None
    assert info["headers"]["host"] == "localhost:%d" % server.server_port
    assert info["headers"]["accept-encoding"] == "identity"
    assert info["client_addr"] == "127.0.0.1"
    assert info["range"] is None


def test_request_info_get_range(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", "/request-info/arg",
                    headers={"range": "bytes=0-99"})
        r = con.getresponse()
        body = r.read()
        assert r.status == http.OK

    info = json.loads(body)
    assert info["range"] == {"first": 0, "last": 99}


def test_request_info_get_range_last_specified(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", "/request-info/arg",
                    headers={"range": "bytes=0-"})
        r = con.getresponse()
        body = r.read()
        assert r.status == http.OK

    info = json.loads(body)
    assert info["range"] == {"first": 0, "last": None}


def test_request_info_get_unsatisfiable_range(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", "/request-info/arg",
                    headers={"range": "bytes=invalid-99"})
        r = con.getresponse()
        r.read()
        assert r.status == http.REQUESTED_RANGE_NOT_SATISFIABLE
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE


def test_request_info_put(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        content = "it works!"
        con.request("PUT", "/request-info/arg", body=content.encode("utf-8"))
        r = con.getresponse()
        assert r.status == http.OK
        info = json.loads(r.read())

    assert info["method"] == "PUT"
    assert info["uri"] == "/request-info/arg"
    assert info["path"] == "/request-info/arg"
    assert info["arg"] == "arg"
    assert info["query"] == {}
    assert info["version"] == "HTTP/1.1"
    assert info["content_length"] == len(content)
    assert info["content"] == content
    assert info["headers"]["accept-encoding"] == "identity"
    assert info["headers"]["content-length"] == str(len(content))
    assert info["content_range"] is None


@pytest.mark.parametrize("content_length", ["not an int", "-1"])
def test_request_invalid_content_length(server, content_length):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request(
            "GET",
            "/request-info/",
            headers={"content-length": content_length})
        r = con.getresponse()
        assert r.status == http.BAD_REQUEST
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE


def test_request_info_put_content_range(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        content = "it works!"
        con.request("PUT", "/request-info/arg",
                    body=content.encode("utf-8"),
                    headers={"content-range": "bytes 0-8/100"})
        r = con.getresponse()
        body = r.read()
        assert r.status == http.OK

    info = json.loads(body)
    assert info["content_range"] == {"first": 0, "last": 8, "complete": 100}


def test_request_info_put_content_range_invalid(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        content = "it works!"
        con.request("PUT", "/request-info/arg",
                    body=content.encode("utf-8"),
                    headers={"content-range": "bytes 0-invalid/*"})
        r = con.getresponse()
        r.read()
        assert r.status == http.BAD_REQUEST
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE


@pytest.mark.parametrize("uri,path,arg", [
    ("/request-info/%d7%90", "/request-info/\u05d0", "\u05d0"),
    ("/request-info%2farg", "/request-info/arg", "arg"),
])
def test_request_info_uri(server, uri, path, arg):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", uri)
        r = con.getresponse()
        assert r.status == http.OK
        info = json.loads(r.read())

    assert info["uri"] == uri
    assert info["path"] == path
    assert info["arg"] == arg


@pytest.mark.parametrize("query_string,parsed_query", [
    # Keep blank values.
    ("a", {"a": ""}),
    # Simple query.
    ("a=1&b=2", {"a": "1", "b": "2"}),
    # Multiple values, last wins.
    ("a=1&a=2", {"a": "2"}),
    # Multiple values, last empty.
    ("a=1&a=2&a", {"a": ""}),
    # Quoted keys and values.
    ("%61=%31", {"a": "1"}),
    # Decoded keys and values {Hebrew Letter Alef: Hebrew Leter Bet}
    # http://unicode.org/charts/PDF/U0590.pdf
    ("%d7%90=%d7%91", {"\u05d0": "\u05d1"}),
])
def test_request_info_query_string(server, query_string, parsed_query):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", "/request-info/?" + query_string)
        r = con.getresponse()
        assert r.status == http.OK
        info = json.loads(r.read())

    assert info["path"] == "/request-info/"
    assert info["query"] == parsed_query


def test_context(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        # No context yet.
        con.request("GET", "/context/this")
        r = con.getresponse()
        assert r.status == http.NOT_FOUND
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE
        r.read()

        # Set value for "this".
        con.request("PUT", "/context/this", body=b"value")
        r = con.getresponse()
        assert r.status == http.OK
        r.read()

        # Should have value now.
        con.request("GET", "/context/this")
        r = con.getresponse()
        assert r.status == http.OK
        assert r.read() == b"value"

        # Remove value for "this".
        con.request("DELETE", "/context/this")
        r = con.getresponse()
        assert r.status == http.NO_CONTENT
        r.read()

        # No context now.
        con.request("GET", "/context/this")
        r = con.getresponse()
        assert r.status == http.NOT_FOUND
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE
        r.read()


def test_context_per_connection(server):
    con1 = http_client.HTTPConnection("localhost", server.server_port)
    con2 = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con1), closing(con2):
        # Set value for "this" in connection 1.
        con1.request("PUT", "/context/this", body=b"con1 value")
        r = con1.getresponse()
        assert r.status == http.OK
        r.read()

        # Connection 1 should have no value.
        con2.request("GET", "/context/this")
        r = con2.getresponse()
        assert r.status == http.NOT_FOUND
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE
        r.read()

        # Set value for "this" in connection 2.
        con2.request("PUT", "/context/this", body=b"con2 value")
        r = con2.getresponse()
        assert r.status == http.OK
        r.read()

        # Connection 1 value did not change.
        con1.request("GET", "/context/this")
        r = con1.getresponse()
        assert r.status == http.OK
        assert r.read() == b"con1 value"


def test_context_deleted_on_close(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        # Set value for "this".
        con.request("PUT", "/context/this", body=b"con value")
        r = con.getresponse()
        assert r.status == http.OK
        r.read()

    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        # Should have no value.
        con.request("GET", "/context/this")
        r = con.getresponse()
        assert r.status == http.NOT_FOUND
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE
        r.read()


def test_context_close(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        # Add a closable object to the connection context
        con.request("PUT", "/close-context/foo")
        r = con.getresponse()
        assert r.status == http.OK
        r.read()

    # Run server thread to detect the close.
    time.sleep(0.1)

    # Closing the connection should close the object.
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", "/close-context/")
        r = con.getresponse()
        assert r.status == http.OK
        log = r.read().decode("utf-8")
        assert "foo was closed" in log


def test_context_close_multiple_objects(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        # Add a closable object to the connection context
        con.request("PUT", "/close-context/foo")
        r = con.getresponse()
        assert r.status == http.OK
        r.read()

        # Add another
        con.request("PUT", "/close-context/bar")
        r = con.getresponse()
        assert r.status == http.OK
        r.read()

    # Run server thread to detect the close.
    time.sleep(0.1)

    # Closing the connection should close both objects.
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", "/close-context/")
        r = con.getresponse()
        assert r.status == http.OK
        log = r.read().decode("utf-8")
        assert "foo was closed" in log
        assert "bar was closed" in log


def test_not_found(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", "/no/such/path")
        r = con.getresponse()
        assert r.status == http.NOT_FOUND
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE


def test_method_not_allowed(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("POST", "/demo/name")
        r = con.getresponse()
        assert r.status == http.METHOD_NOT_ALLOWED
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE


def test_invalid_method(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("FOO", "/demo/name")
        r = con.getresponse()
        assert r.status == http.METHOD_NOT_ALLOWED
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE


def test_client_error_get(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", "/client-error/")
        r = con.getresponse()
        assert r.status == http.FORBIDDEN
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE


def test_client_error_put(server):
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        try:
            con.request("PUT", "/client-error/", body=b"x" * 1024**2)
        except socket.error as e:
            if e.args[0] not in (errno.EPIPE, errno.ESHUTDOWN):
                raise
        r = con.getresponse()
        assert r.status == http.FORBIDDEN
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE


def test_internal_error_get(server):
    # Internal error should not expose secret data in client response.
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", "/server-error/")
        r = con.getresponse()
        assert r.status == http.INTERNAL_SERVER_ERROR
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE
        assert "secret" not in r.read().decode("utf-8")


def test_internal_error_put(server):
    # Internal error should not expose secret data in client response.
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        try:
            con.request("PUT", "/server-error/", body=b"x" * 1024**2)
        except socket.error as e:
            if e.args[0] not in (errno.EPIPE, errno.ESHUTDOWN):
                raise
        r = con.getresponse()
        assert r.status == http.INTERNAL_SERVER_ERROR
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE
        assert "secret" not in r.read().decode("utf-8")


def test_server_socket_error_get(server):
    # Socket error on server side should report as INTERNAL_SERVER_ERROR.
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        con.request("GET", "/server-socket-error/")
        r = con.getresponse()
        assert r.status == http.INTERNAL_SERVER_ERROR
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE


def test_server_socket_error_put(server):
    # Socket error on server side should report as INTERNAL_SERVER_ERROR.
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        try:
            con.request("PUT", "/server-socket-error/", body=b"x" * 1024**2)
        except socket.error as e:
            if e.args[0] not in (errno.EPIPE, errno.ESHUTDOWN):
                raise
        r = con.getresponse()
        assert r.status == http.INTERNAL_SERVER_ERROR
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE


@pytest.mark.parametrize("data", [None, b"", b"read me"])
def test_keep_connection_on_error(server, data):
    # When a request does not have a payload, the server can keep the
    # connection open after and error.
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        # Disabling auto_open so we can test if a connection was closed.
        con.auto_open = False
        con.connect()

        # Send couple of requests - all should fail, without closing the
        # connection.
        for i in range(3):
            con.request("PUT", "/keep-connection/", body=data)
            r = con.getresponse()
            r.read()
            assert r.status == http.FORBIDDEN
            assert r.getheader("content-type") == ERROR_CONTENT_TYPE


def test_close_connection_on_error(server):
    # When payload was not read completely, the server must close the
    # connection.
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        # Disabling auto_open so we can test if a connection was closed.
        con.auto_open = False
        con.connect()

        # Send a request - it should fail without reading the request,
        # so the server will close the connection.
        con.request("PUT", "/client-error/", body=b"read me")
        r = con.getresponse()
        r.read()
        assert r.status == http.FORBIDDEN
        assert r.getheader("content-type") == ERROR_CONTENT_TYPE

        # Try to send another request. This will fail since we disabled
        # auto_open.  Fails in request() or in getresponse(), probably
        # depends on timing.
        with pytest.raises(
                (http_client.NotConnected, http_client.BadStatusLine)):
            con.request("GET", "/client-error/")
            con.getresponse()


def test_close_connection_on_partial_response(server):
    # When the response was not sent completely, the server must close
    # the connection.
    con = http_client.HTTPConnection("localhost", server.server_port)
    with closing(con):
        # Disabling auto_open so we can test if a connection was closed.
        con.auto_open = False
        con.connect()

        # Send a request - it should failed because the server
        # closed the connection before sending the entire response.
        con.request("GET", "/partial-response/")
        r = con.getresponse()
        assert r.status == http.OK
        with pytest.raises(http_client.IncompleteRead):
            r.read()

        # Try to send another request. This will fail since we disabled
        # auto_open.  Fails in request() or in getresponse(), probably
        # depends on timing.
        with pytest.raises(
                (http_client.NotConnected, http_client.BadStatusLine)):
            con.request("GET", "/client-error/")
            con.getresponse()


@pytest.mark.parametrize("header,first,last", [
    # Both first and last
    ("bytes=0-99", 0, 99),
    # One byte range.
    ("bytes=42-42", 42, 42),
    # From 99 to end.
    ("bytes=99-", 99, None),
    # Last 99 byts
    ("bytes=-99", -99, None),
])
def test_range_parse(header, first, last):
    r = http.Range.parse(header)
    assert r.first == first
    assert r.last is last


@pytest.mark.parametrize("header", [
    # Missing bytes
    "cats=0-99",
    # Wrong case
    "BYTES=0-99",
    # Extra spaces
    "bytes =0-99",
    "bytes= 0-99",
    "bytes=0 -99",
    "bytes=0- 99",
    "bytes=0-99 ",
    # Missing =
    "bytes 0-99",
    # Missing -
    "bytes=99",
    # first > last
    "bytes=99-98",
    # negative first
    "bytes=-42-",
    # first and negative last (conflict)
    "bytes=42--99",
    # ultiple ranges not supported yet.
    "bytes=0-499,500-599",
])
def test_range_parse_not_satisfiable(header):
    with pytest.raises(http.Error) as e:
        http.Range.parse(header)
    assert e.value.code == http.REQUESTED_RANGE_NOT_SATISFIABLE


@pytest.mark.parametrize("header,first,last,complete", [
    # First 100 bytes of 200 bytes.
    ("bytes 0-99/200", 0, 99, 200),
    # First byte of 200 bytes.
    ("bytes 0-0/200", 0, 0, 200),
    # Last byte of 200 bytes.
    ("bytes 199-199/200", 199, 199, 200),
    # Last unspeficied
    ("bytes 100-*/200", 100, None, 200),
    # Complete unspecified
    ("bytes 100-199/*", 100, 199, None),
    # Last and complete unspecified
    ("bytes 100-*/*", 100, None, None),
])
def test_content_range_parse(header, first, last, complete):
    r = http.ContentRange.parse(header)
    assert r.first == first
    assert r.last is last
    assert r.complete is complete


@pytest.mark.parametrize("header", [
    # Unsupported unit
    "cats 0-99/200",
    # Missing element
    "bytes 0-99/",
    "bytes 0-99",
    "bytes -99/200",
    "bytes 99-/200",
    "bytes 99/200",
    "bytes /200",
    # Invalid numberic values
    "bytes invalid-99/200",
    "bytes 0-invalid/200",
    "bytes 0-99/invalid",
    # Unspecified first
    "bytes *-99/200",
    # first > last
    "bytes 99-98/200",
    # last >= complete
    "bytes 0-200/200",
    "bytes 0-201/200",
    # Unsatisfied range (valid, but unsupported)
    "bytes */200",
])
def test_content_range_parse_invalid(header):
    with pytest.raises(http.Error) as e:
        http.ContentRange.parse(header)
    assert e.value.code == http.BAD_REQUEST


@pytest.mark.parametrize("listen_address, expected_server_address", [
    ("localhost", "::1"),
    ("127.0.0.1", "127.0.0.1"),
    ("::1", "::1"),
])
def test_server_bind_loopback(listen_address, expected_server_address):
    with demo_server(listen_address) as server:
        assert server.server_address[0] == expected_server_address
        # TCPServer returns tuple of different length for IPv4 and IPv6. We
        # change server_address in Server.server_bind() to be always tuple
        # of hostname and port. Check it here, that it has always two elements.
        assert len(server.server_address) == 2

        con = http_client.HTTPConnection("localhost", server.server_port)
        with closing(con):
            con.request("GET", "/demo/name")
            con.getresponse()


def test_server_bind_ipv4():
    with demo_server("") as server:
        # Connecting via IPv4 address should work.
        con = http_client.HTTPConnection("127.0.0.1", server.server_port)
        with closing(con):
            con.request("GET", "/demo/name")
            con.getresponse()

        # Connecting via IPv6 address should fail.
        con = http_client.HTTPConnection("::1", server.server_port)
        with closing(con):
            with pytest.raises(ConnectionRefusedError):
                con.request("GET", "/demo/name")


@pytest.mark.parametrize("connect_address", [
    "127.0.0.1",
    "::1",
    "localhost",
    socket.gethostname(),
])
def test_server_bind_dual_stack(connect_address):
    with demo_server("::") as server:
        con = http_client.HTTPConnection(connect_address, server.server_port)
        with closing(con):
            con.request("GET", "/demo/name")
            con.getresponse()


@pytest.mark.parametrize(
    "address", [a[4][0] for a in http.find_addresses(socket.gethostname())])
def test_server_bind_addr_from_hostname(address):
    with demo_server(address) as server:
        con = http_client.HTTPConnection(address, server.server_port)
        with closing(con):
            con.request("GET", "/demo/name")
            con.getresponse()


@pytest.mark.parametrize("listen_address", [
    "localhost",
    "127.0.0.1",
])
def test_prefer_ipv4(listen_address):
    with demo_server(listen_address, prefer_ipv4=True) as server:
        assert server.server_address[0] == "127.0.0.1"

        con = http_client.HTTPConnection(listen_address, server.server_port)
        with closing(con):
            con.request("GET", "/demo/name")
            con.getresponse()



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest
import json

from ovirt_imageio import client
from ovirt_imageio._internal.units import MiB, GiB


class FakeTime:

    def __init__(self):
        self.now = 0.0

    def __call__(self):
        return self.now


class FakeFile:

    def __init__(self):
        self.last = None

    def write(self, s):
        self.last = s

    def flush(self):
        pass


def test_draw_text():
    fake_time = FakeTime()
    f = FakeFile()

    # Size is unknown at this point.
    pb = client.ProgressBar(phase="setting up", output=f, now=fake_time)
    line = "[ ---- ] 0 bytes, 0.00 s, 0 bytes/s | setting up"
    assert f.last == line.ljust(79) + "\r"

    # Size was updated, but no bytes were transferred yet.
    fake_time.now += 0.1
    pb.size = 3 * GiB
    line = "[   0% ] 0 bytes, 0.10 s, 0 bytes/s | setting up"
    assert f.last == line.ljust(79) + "\r"

    # Phase was updated.
    fake_time.now += 0.2
    pb.phase = "downloading image"
    line = "[   0% ] 0 bytes, 0.30 s, 0 bytes/s | downloading image"
    assert f.last == line.ljust(79) + "\r"

    # Write some data...
    fake_time.now += 0.8
    pb.update(512 * MiB)
    line = "[  16% ] 512.00 MiB, 1.10 s, 465.45 MiB/s | downloading image"
    assert f.last == line.ljust(79) + "\r"

    # Write zeros (much faster)...
    fake_time.now += 0.2
    pb.update(2 * GiB)
    line = "[  83% ] 2.50 GiB, 1.30 s, 1.92 GiB/s | downloading image"
    assert f.last == line.ljust(79) + "\r"

    # More data, slow down again...
    fake_time.now += 1.0
    pb.update(512 * MiB)
    line = "[ 100% ] 3.00 GiB, 2.30 s, 1.30 GiB/s | downloading image"
    assert f.last == line.ljust(79) + "\r"

    # Cleaning up after download.
    pb.phase = "cleaning up"
    line = "[ 100% ] 3.00 GiB, 2.30 s, 1.30 GiB/s | cleaning up"
    assert f.last == line.ljust(79) + "\r"

    # Cleaning can take few seconds, lowwing the the rate.
    fake_time.now += 3.0
    pb.phase = "download completed"
    line = "[ 100% ] 3.00 GiB, 5.30 s, 579.62 MiB/s | download completed"
    assert f.last == line.ljust(79) + "\r"

    # Closing prints the final line with a newline terminator.
    pb.close()
    line = "[ 100% ] 3.00 GiB, 5.30 s, 579.62 MiB/s | download completed"
    assert f.last == line.ljust(79) + "\n"


def test_draw_json():
    fake_time = FakeTime()
    f = FakeFile()

    # Size is unknown at this point.
    pb = client.ProgressBar(
        phase="setting up", output=f, format="json", now=fake_time)
    assert json.loads(f.last) == {
        "transferred": 0,
        "elapsed": 0.0,
        "description": "setting up",
    }

    # Size was updated, but no bytes were transferred yet.
    fake_time.now += 0.1
    pb.size = 3 * GiB
    assert json.loads(f.last) == {
        "transferred": 0,
        "elapsed": fake_time.now,
        "description": "setting up",
        "size": 3 * GiB,
    }

    # Phase was updated.
    fake_time.now += 0.2
    pb.phase = "downloading image"
    assert json.loads(f.last) == {
        "transferred": 0,
        "elapsed": fake_time.now,
        "description": "downloading image",
        "size": 3 * GiB,
    }

    # All data transferred.
    fake_time.now += 2.0
    pb.update(3 * GiB)
    assert json.loads(f.last) == {
        "transferred": 3 * GiB,
        "elapsed": fake_time.now,
        "description": "downloading image",
        "size": 3 * GiB,
    }

    # Cleaning up after download.
    pb.phase = "cleaning up"
    assert json.loads(f.last) == {
        "transferred": 3 * GiB,
        "elapsed": fake_time.now,
        "description": "cleaning up",
        "size": 3 * GiB,
    }

    # Closing prints the final line.
    pb.close()
    assert json.loads(f.last) == {
        "transferred": 3 * GiB,
        "elapsed": fake_time.now,
        "description": "cleaning up",
        "size": 3 * GiB,
    }


def test_with_size():
    fake_time = FakeTime()
    f = FakeFile()

    client.ProgressBar(phase="starting", size=3 * GiB, output=f, now=fake_time)
    line = "[   0% ] 0 bytes, 0.00 s, 0 bytes/s | starting"
    assert f.last == line.ljust(79) + "\r"


def test_without_phase():
    fake_time = FakeTime()
    f = FakeFile()

    client.ProgressBar(output=f, now=fake_time)
    line = "[ ---- ] 0 bytes, 0.00 s, 0 bytes/s"
    assert f.last == line.ljust(79) + "\r"


def test_close():
    f = FakeFile()
    pb = client.ProgressBar(output=f)
    pb.size = 1 * GiB
    pb.update(512 * MiB)
    pb.close()
    f.last = None

    # Once closed, update does not redraw.
    pb.update(512 * MiB)
    assert f.last is None

    # Changing size does nothing.
    pb.size = 2 * GiB
    assert f.last is None
    assert pb.size == 1 * GiB

    # Changing phase does nothing.
    pb.phase = "new phase"
    assert f.last is None
    assert pb.phase is None

    # Closing twice does not redraw.
    pb.close()
    assert f.last is None


def test_contextmanager():
    f = FakeFile()
    with client.ProgressBar(size=GiB, output=f) as pb:
        pb.update(GiB)
        assert f.last.endswith("\r")

    assert f.last.endswith("\n")


def test_error_phase_default():
    f = FakeFile()
    pb = client.ProgressBar(phase="running command", size=GiB, output=f)
    with pytest.raises(RuntimeError):
        with pb:
            raise RuntimeError

    assert f.last.endswith("\n")
    assert pb.phase == "command failed"


def test_error_phase_custom():
    f = FakeFile()
    pb = client.ProgressBar(
        phase="starting operation",
        error_phase="operation failed",
        size=GiB,
        output=f)
    with pytest.raises(RuntimeError):
        with pb:
            raise RuntimeError

    assert f.last.endswith("\n")
    assert pb.phase == "operation failed"



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import collections
import time

from contextlib import contextmanager

from . import util


class Clock:
    """
    Measure time for complex flows.

    This clock is useful for timing complex flows, when you want to record
    multiple timings for a single flow. For example, the total time, and the
    time of each step in the flow.

    This is similar to MoinMoin.util.clock.Clock:
    https://bitbucket.org/thomaswaldmann/moin-2.0/src/

    And vdsm.common.time.Clock:
    https://github.com/oVirt/vdsm/blob/master/lib/vdsm/common/time.py#L45

    Usage::

        clock = time.Clock()
        ...
        clock.start("total")
        ...
        clock.start("read")
        clock.stop("read")
        clock.start("write")
        clock.stop("write")
        clock.start("read")
        clock.stop("read")
        clock.start("write")
        clock.stop("write")
        ...
        clock.start("sync")
        clock.stop("sync")
        ...
        clock.stop("total")
        log.info("times=%s", clock)

    """

    def __init__(self, now=time.monotonic):
        # Keep insertion order for nicer output in __repr__.
        self._stats = collections.OrderedDict()
        self._now = now

    def start(self, name):
        s = self._stats.get(name)
        if s is None:
            s = self._stats[name] = Stats(name)

        if s.started is not None:
            raise RuntimeError("Stats %r was already started" % name)

        s.started = self._now()

        return s

    def stop(self, name):
        s = self._lookup_started(name)
        return self._stop(s, True)

    def abort(self, name):
        s = self._lookup_started(name)
        return self._stop(s, False)

    @contextmanager
    def run(self, name):
        s = self.start(name)
        try:
            yield s
        except BaseException:
            self._stop(s, False)
            raise
        else:
            self._stop(s, True)

    def _lookup_started(self, name):
        s = self._stats.get(name)
        if s is None:
            raise RuntimeError("No such stats %r" % name)

        if s.started is None:
            raise RuntimeError("Stats %r was not started" % name)

        return s

    def _stop(self, s, completed):
        elapsed = self._now() - s.started
        s.seconds += elapsed
        s.started = None
        if completed:
            s.ops += 1

        return elapsed

    def __repr__(self):
        now = self._now()
        stats = []
        for s in self._stats.values():
            if s.started is not None:
                seconds = now - s.started
            else:
                seconds = s.seconds
            values = [
                "{} ops".format(s.ops),
                "{:.6f} s".format(seconds),
            ]
            if s.bytes:
                values.append(util.humansize(s.bytes))
                values.append(util.humansize(s.bytes / seconds) + "/s")

            stats.append("[{} {}]".format(s.name, ", ".join(values)))

        return " ".join(stats)


class NullClock:
    """
    Clock that does nothing.

    This avoids checking for None clock, so users can do:

        with clock.run("name"):
            stuff to measure...

    Even if timing is disabled.
    """

    def start(self, name):
        pass

    def stop(self, name):
        return 0

    @contextmanager
    def run(self, name):
        yield _NULL_STATS

    def __repr__(self):
        return ""


class Stats:

    def __init__(self, name):
        self.name = name
        self.seconds = 0.0
        self.ops = 0
        self.bytes = 0
        self.started = None


class NullStats:

    @property
    def bytes(self):
        return 0

    @bytes.setter
    def bytes(self, value):
        pass


_NULL_STATS = NullStats()



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest

from . import ci
from . import testutil

requires_ipv6 = pytest.mark.skipif(
    not testutil.ipv6_enabled(),
    reason="IPv6 not available")

flaky_in_ovirt_ci = pytest.mark.xfail(
    ci.is_ovirt(), reason="Test is flaky in oVirt CI", strict=False)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import io
import logging
import os
import subprocess
import urllib.parse

from collections import namedtuple

import pytest
import userstorage

from ovirt_imageio._internal import nbd
from ovirt_imageio._internal import qemu_nbd
from ovirt_imageio._internal import util

# Mark tests with xfail if userstorage is missing
userstorage.missing_handler = pytest.xfail

log = logging.getLogger("test")


@pytest.fixture
def tmpfile(tmpdir):
    """
    Return a path to an empty temporary file.
    """
    f = tmpdir.join("tmpfile")
    f.write("")
    return str(f)


@pytest.fixture
def tmpurl(tmpfile):
    """
    Return a file: url to an empty temporary file.
    """
    return urllib.parse.urlparse("file:" + tmpfile)


@pytest.fixture
def nbd_server(tmpdir):
    """
    Returns nbd_server exporting a temporary file.

    The test may configure the server before starting it. Typical example is
    setting the read_only flag to create a read only server.

    The test must start the server. The test framework will stop the server
    when the test ends.
    """
    image = str(tmpdir.join("image"))
    with io.open(image, "wb") as f:
        f.truncate(10 * 1024**2)

    sock = nbd.UnixAddress(tmpdir.join("sock"))

    server = qemu_nbd.Server(image, "raw", sock)
    try:
        yield server
    finally:
        server.stop()


@pytest.fixture
def base_image():
    path = "/var/tmp/imageio-images/cirros-0.3.5.img"
    if not os.path.exists(path):
        pytest.skip(f"{path} not available")

    return path


# Arguments to ssl.{server,client}_context()
PKI = namedtuple("PKI", "cafile,certfile,keyfile")


@pytest.fixture(scope="session")
def tmp_pki(tmpdir_factory):
    tmpdir = tmpdir_factory.mktemp("pki")
    certfile = str(tmpdir.join("cert.pem"))
    keyfile = str(tmpdir.join("key.pem"))

    cmd = [
        "openssl", "req",
        "-new",
        "-x509",
        "-nodes",
        "-batch",
        "-days", "2",
        "-subj", "/CN=localhost",
        "-out", certfile,
        "-keyout", keyfile,
    ]

    subprocess.check_output(cmd)

    return PKI(certfile, certfile, keyfile)


class FakeTime:

    def __init__(self):
        self.now = 0

    def monotonic_time(self):
        return self.now


@pytest.fixture
def fake_time(monkeypatch):
    """
    Monkeypatch util.monotonic_time for testing time related operations.

    Returns FakeTime instance. Modifying instance.now change the value returned
    from the monkeypatched util.monotonic_time().
    """
    time = FakeTime()
    monkeypatch.setattr(util, "monotonic_time", time.monotonic_time)
    return time



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest

from ovirt_imageio._internal import auth
from ovirt_imageio._internal import config
from ovirt_imageio._internal import errors
from ovirt_imageio._internal import services


@pytest.mark.parametrize("port", [-1, 65536])
def test_invalid_remote_port(port):
    cfg = config.load(["test/conf/daemon.conf"])
    authorizer = auth.Authorizer(cfg)
    cfg.remote.port = port
    with pytest.raises(errors.InvalidConfig):
        services.RemoteService(cfg, authorizer)


@pytest.mark.parametrize("port", [-1, 65536])
def test_invalid_control_port(port):
    cfg = config.load(["test/conf/proxy.conf"])
    authorizer = auth.Authorizer(cfg)
    cfg.control.port = port
    with pytest.raises(errors.InvalidConfig):
        services.ControlService(cfg, authorizer)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import json
import logging

from .. import errors
from .. import http
from .. import validate

log = logging.getLogger("tickets")


class Handler:
    """
    Handle requests for the /tickets/ resource.
    """

    def __init__(self, config, auth):
        self.config = config
        self.auth = auth

    def get(self, req, resp, ticket_id):
        if not ticket_id:
            raise http.Error(http.BAD_REQUEST, "Ticket id is required")

        try:
            ticket = self.auth.get(ticket_id)
        except KeyError:
            raise http.Error(
                http.NOT_FOUND, "No such ticket {!r}".format(ticket_id))

        ticket_info = ticket.info()
        log.debug("[%s] GET transfer=%s", req.client_addr, ticket.transfer_id)
        resp.send_json(ticket_info)

    def put(self, req, resp, ticket_id):
        if not ticket_id:
            raise http.Error(http.BAD_REQUEST, "Ticket id is required")

        # TODO: Reject too big ticket json. We know the size of a ticket based
        # on the size of the keys and values.
        try:
            ticket_dict = json.loads(req.read())
        except ValueError as e:
            raise http.Error(
                http.BAD_REQUEST,
                "Ticket is not in a json format: {}".format(e))

        log.info("[%s] ADD transfer=%s", req.client_addr, ticket_dict)
        try:
            self.auth.add(ticket_dict)
        except errors.InvalidTicket as e:
            raise http.Error(
                http.BAD_REQUEST, "Invalid ticket: {}".format(e))

    def patch(self, req, resp, ticket_id):
        if not ticket_id:
            raise http.Error(http.BAD_REQUEST, "Ticket id is required")

        # TODO: Reject requests with too big payload. We know the size of a
        # ticket patch message based on the keys and values.
        try:
            patch = json.loads(req.read())
        except ValueError as e:
            raise http.Error(
                http.BAD_REQUEST, "Invalid patch: {}".format(e))

        timeout = validate.integer(patch, "timeout", minval=0)

        try:
            ticket = self.auth.get(ticket_id)
        except KeyError:
            raise http.Error(
                http.NOT_FOUND, "No such ticket: {}".format(ticket_id))

        log.info("[%s] EXTEND timeout=%s transfer=%s",
                 req.client_addr, timeout, ticket.transfer_id)
        ticket.extend(timeout)

    def delete(self, req, resp, ticket_id):
        """
        Delete a ticket if exists.

        Note that DELETE is idempotent;  the client can issue multiple DELETE
        requests in case of network failures. See
        https://tools.ietf.org/html/rfc7231#section-4.2.2.
        """
        if ticket_id:
            try:
                ticket = self.auth.get(ticket_id)
            except KeyError:
                # Ticket was already deleted.
                log.info("[%s] REMOVE transfer=(deleted)", req.client_addr)
            else:
                log.info("[%s] REMOVE transfer=%s",
                         req.client_addr, ticket.transfer_id)
                try:
                    self.auth.remove(ticket_id)
                except errors.TicketCancelTimeout as e:
                    # The ticket is still used by some connection so we cannot
                    # remove it. The caller can retry the call again when the
                    # number connections reach zero.
                    raise http.Error(http.CONFLICT, str(e))
        else:
            log.info("[%s] REMOVE all tickets", req.client_addr)
            self.auth.clear()

        resp.status_code = http.NO_CONTENT



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Helpers for handling issues on specific CI environment.
"""

import os


def is_ovirt():
    return "OVIRT_CI" in os.environ


def is_travis():
    return "TRAVIS_CI" in os.environ



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Tool for transferring disk images.
"""

import logging
import signal
import sys

from . import _app
from . import _download
from . import _options
from . import _upload

log = logging.getLogger("tool")


def main():
    parser = _options.Parser()
    _download.register(parser)
    _upload.register(parser)
    args = parser.parse()

    logging.basicConfig(
        filename=args.log_file,
        level=args.log_level.upper(),
        format="%(asctime)s %(levelname)-7s (%(threadName)s) [%(name)s] "
               "%(message)s")

    _app.setup_signals()

    try:
        args.command(args)
    except _app.TerminatedBySignal as e:
        # SIGINT is likey result of Control+C in the shell.
        level = logging.INFO if e.signal == signal.SIGINT else logging.ERROR
        log.log(level, "Exiting: %s", e)
        sys.exit(128 + e.signal)
    except Exception:
        log.exception("Command failed")
        sys.exit(1)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
http helpers.
"""

import errno
import http.client
import logging

from pprint import pprint

from ovirt_imageio._internal import errors
from ovirt_imageio._internal import ssl
from ovirt_imageio._internal import uhttp

log = logging.getLogger("test")


class HTTPClient:

    def __init__(self, con):
        self.con = con

    def get(self, uri, headers=None):
        return self.request("GET", uri, headers=headers)

    def put(self, uri, body, headers=None):
        return self.request("PUT", uri, body=body, headers=headers)

    def delete(self, uri, headers=None):
        return self.request("DELETE", uri, headers=headers)

    def patch(self, uri, body, headers=None):
        return self.request("PATCH", uri, body=body, headers=headers)

    def options(self, uri, headers=None):
        return self.request("OPTIONS", uri, headers=headers)

    def request(self, method, uri, body=None, headers=None):
        try:
            self.con.request(method, uri, body=body, headers=headers or {})
        except EnvironmentError as e:
            if not (e.errno == errno.EPIPE and body):
                raise
            log.warning("Error sending request: %s", e)
        return response(self.con)

    def raw_request(self, method, uri, body=None, headers=None):
        """
        Use this to send bad requests - this will send only the headers set in
        headers, no attempt is made to create a correct request.
        """
        self.con.putrequest(method, uri)
        if headers:
            for name, value in headers.items():
                self.con.putheader(name, value)
        self.con.endheaders()

        try:
            if body:
                self.con.send(body)
        except EnvironmentError as e:
            if e.errno != errno.EPIPE:
                raise
            log.warning("Error sending body: %s", e)

        return response(self.con)

    def close(self):
        self.con.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


class RemoteClient(HTTPClient):

    def __init__(self, cfg):
        if cfg.tls.enable:
            context = ssl.client_context(cfg.tls.ca_file)
            con = http.client.HTTPSConnection(
                cfg.remote.host,
                cfg.remote.port,
                context=context)
        else:
            con = http.client.HTTPConnection(cfg.remote.host, cfg.remote.port)

        super().__init__(con)


class LocalClient(HTTPClient):

    def __init__(self, cfg):
        super().__init__(uhttp.UnixHTTPConnection(cfg.local.socket))


class ControlClient(HTTPClient):

    def __init__(self, cfg):
        transport = cfg.control.transport.lower()
        if transport == "tcp":
            con = http.client.HTTPConnection("localhost", cfg.control.port)
        elif transport == "unix":
            con = uhttp.UnixHTTPConnection(cfg.control.socket)
        else:
            raise errors.InvalidConfig("control.transport", transport)

        super().__init__(con)


def response(con):
    res = con.getresponse()
    pprint((res.status, res.reason, res.version))
    pprint(res.getheaders())
    return res



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

from ovirt_imageio._internal import errors


def test_str():
    e = errors.PartialContent(50, 42)
    assert str(e) == "Requested 50 bytes, available 42 bytes"



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import json
import logging
import socket
import subprocess

from contextlib import closing
from uuid import uuid4

log = logging.getLogger("test")


def head(b):
    return b[:10]


def random_tcp_port():
    """
    Find a random (likely) unused port.

    The port is unused when the call return, but another process may
    grab it.  If you don't control the environmemnt, be prepared for
    bind failures.
    """
    s = socket.socket()
    with closing(s):
        s.bind(("localhost", 0))
        port = s.getsockname()[1]
        log.debug("Found unused TCP port %s", port)
        return port


def create_ticket(uuid=None, ops=None, timeout=300, size=2**64,
                  url="file:///tmp/foo.img", transfer_id=None, filename=None,
                  sparse=None, dirty=None, inactivity_timeout=120):
    d = {
        "uuid": uuid or str(uuid4()),
        "timeout": timeout,
        "ops": ["read", "write"] if ops is None else ops,
        "size": size,
        "url": url,
        "transfer_id": transfer_id or str(uuid4()),
    }
    if filename is not None:
        d["filename"] = filename
    if sparse is not None:
        d["sparse"] = sparse
    if dirty is not None:
        d["dirty"] = dirty
    if inactivity_timeout is not None:
        d["inactivity_timeout"] = inactivity_timeout
    return d


def create_tempfile(tmpdir, name, data=b'', size=None):
    file = tmpdir.join(name)
    with open(str(file), 'wb') as f:
        if size is not None:
            f.truncate(size)
        if data:
            f.write(data)
    return file


def ipv6_enabled():
    out = subprocess.check_output(["ip", "-6", "-j", "addr"])
    addresses = json.loads(out)
    return len(addresses) > 0



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Client application global state.
"""

import signal

_termination_signal = None
_registered_signals = False


class TerminatedBySignal(Exception):

    def __init__(self, signo):
        self.signal = signo

    def __str__(self):
        return f"Terminated by signal {self.signal}"


def check_terminated():
    """
    Raise TerminatedBySignal if termination signal was received.
    """
    if _termination_signal is not None:
        raise TerminatedBySignal(_termination_signal)


def setup_signals():
    global _registered_signals
    signal.signal(signal.SIGINT, _handle_signal)
    signal.signal(signal.SIGTERM, _handle_signal)
    _registered_signals = True


def is_handling_signals():
    """
    Return True if the application registered signals handlers.
    """
    return _registered_signals


def _handle_signal(signo, frame):
    global _termination_signal
    if _termination_signal is None:
        _termination_signal = signo



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
_internal - imageio internal moudules used by the server and client.

Do not use modules form this package; they are internal implemention details
and may change on each release in incompatible way.
"""



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import socket

from contextlib import closing

from . import testutil


def test_random_tcp_port():
    # Use 100 iterations to detect flakyness early.
    for i in range(100):
        s = socket.socket()
        with closing(s):
            port = testutil.random_tcp_port()
            s.bind(("localhost", port))


def test_create_tempfile_hole(tmpdir):
    size = 1024
    file = testutil.create_tempfile(tmpdir, "image", size=size)
    assert file.read() == "\0" * size


def test_create_tempfile_data(tmpdir):
    size = 1024
    file = testutil.create_tempfile(tmpdir, "image", data=b"x" * size)
    assert file.read() == "x" * size


def test_create_tempfile_data_and_size(tmpdir):
    data_size = 512
    virtual_size = 1024
    file = testutil.create_tempfile(
        tmpdir, "image", data=b"x" * data_size, size=virtual_size)
    assert file.read() == "x" * data_size + "\0" * (virtual_size - data_size)



#!/venv/bin/python3
#
# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import argparse
import json
import os
import subprocess

from ovirt_imageio._internal import qemu_img
from ovirt_imageio._internal import qemu_nbd
from ovirt_imageio._internal import nbd
from ovirt_imageio._internal import util

TICKET_PATH = "/app/ticket.json"
OVIRT_IMAGEIO = "/venv/bin/ovirt-imageio"


def main():
    args = _parse_args()
    image_info = qemu_img.info(args.image)
    with util.tmp_dir("imageio-") as base:
        socket_path = os.path.join(base, "sock")

        ticket_info = _generate_ticket(
            args.ticket_id, socket_path, image_info["virtual-size"])
        with open(TICKET_PATH, "w") as f:
            f.write(json.dumps(ticket_info))

        sock = nbd.UnixAddress(socket_path)
        with qemu_nbd.run(args.image, image_info["format"], sock):
            subprocess.run(
                [OVIRT_IMAGEIO, "--ticket", TICKET_PATH], check=True)


def _generate_ticket(uuid, socket, size):
    ticket_info = {}
    ticket_info["uuid"] = uuid
    ticket_info["url"] = f"nbd:unix:{socket}"
    ticket_info["size"] = size
    ticket_info["timeout"] = 3000
    ticket_info["inactivity_timeout"] = 300
    ticket_info["ops"] = ["read"]
    return ticket_info


def _parse_args():
    parser = argparse.ArgumentParser(description="ovirt-imageio")
    parser.add_argument("image", help="Path to image.")
    parser.add_argument(
        "--ticket-id",
        help="Optional. Set the ID of the ticket. It is recommended "
        "to use with random string for better security. "
        "Defaults to the image filename.")

    args = parser.parse_args()
    # Ticket ID defaults to image filename if not set.
    args.ticket_id = args.ticket_id or os.path.basename(args.image)
    return args


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print("Interrupted, shutting down")



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
import pprint

from ovirt_imageio._internal import nbd
from ovirt_imageio._internal import qemu_img

from . import qemu
from . import qmp

log = logging.getLogger("test")


def test_query_status(tmpdir):
    # Simplest possible test.
    image = str(tmpdir.join("image.raw"))
    with open(image, "wb") as f:
        f.truncate(1024**2)

    qmp_sock = nbd.UnixAddress(tmpdir.join("qmp.sock"))

    with qemu.run(image, "raw", qmp_sock, start_cpu=False):
        with qmp.Client(qmp_sock) as c:
            r = c.execute("query-status")
            assert r["status"] == "prelaunch"


def test_add_bitmap(tmpdir):
    # Test command with arguments. This is also interesting for incremental
    # backup flows.
    image = str(tmpdir.join("image.qcow2"))
    qemu_img.create(image, "qcow2", size=1024**3)

    qmp_sock = nbd.UnixAddress(tmpdir.join("qmp.sock"))

    with qemu.run(image, "qcow2", qmp_sock, start_cpu=False):
        with qmp.Client(qmp_sock) as c:
            c.execute("block-dirty-bitmap-add", {
                "node": "file0",
                "name": "bitmap0",
            })
            node = qmp.find_node(c, image)
            log.debug("Found node:\n%s", pprint.pformat(node))

            if qemu.version() >= (6, 0, 0):
                bitmaps = node["inserted"]["dirty-bitmaps"]
            else:
                bitmaps = node["dirty-bitmaps"]
            assert bitmaps[0]["name"] == "bitmap0"



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import os
from contextlib import closing

import pytest
import userstorage

from ovirt_imageio._internal import errors
from ovirt_imageio._internal import extent
from ovirt_imageio._internal import qemu_img
from ovirt_imageio._internal import util
from ovirt_imageio._internal.backends import nbd

from .. import storage

BACKENDS = userstorage.load_config("storage.py").BACKENDS


@pytest.fixture(
    params=[
        BACKENDS["file-512-ext4"],
        BACKENDS["file-512-xfs"],
        BACKENDS["file-4k-ext4"],
        BACKENDS["file-4k-xfs"],
    ],
    ids=str
)
def user_file(request):
    backend = request.param
    with backend:
        yield storage.Backend(backend)


def test_debugging_interface(nbd_server):
    nbd_server.start()
    with nbd.open(nbd_server.url, "r+") as b:
        assert b.readable()
        assert b.writable()
        assert not b.sparse
        assert b.name == "nbd"


def test_open_read_write(nbd_server):
    nbd_server.start()
    with nbd.open(nbd_server.url, "r+") as b:
        assert b.readable()
        assert b.writable()

        data = b"data"
        b.write(data)
        assert b.tell() == len(data)

        b.zero(4)
        size = len(data) + 4
        assert b.tell() == size

        with closing(util.aligned_buffer(size)) as buf:
            b.seek(0)
            assert b.readinto(buf) == size
            assert buf[:] == data + b"\0" * 4
        b.flush()


def test_open_readonly(nbd_server):
    nbd_server.read_only = True
    nbd_server.start()
    with nbd.open(nbd_server.url) as b:
        assert b.readable()
        assert not b.writable()

        with pytest.raises(IOError):
            b.write(b"data")
        assert b.tell() == 0

        with pytest.raises(IOError):
            b.zero(4)
        assert b.tell() == 0

        with closing(util.aligned_buffer(100)) as buf:
            buf.write(b"x" * 100)
            assert b.readinto(buf) == len(buf)
            assert buf[:] == b"\0" * len(buf)

        b.flush()


def test_open_writeonly(nbd_server):
    nbd_server.start()
    with nbd.open(nbd_server.url, "w") as b:
        assert not b.readable()
        assert b.writable()

        data = b"data"
        b.write(data)
        assert b.tell() == len(data)

        with pytest.raises(IOError):
            with closing(util.aligned_buffer(100)) as buf:
                b.readinto(buf)

        b.flush()


def test_invalid_mode(nbd_server):
    nbd_server.start()
    with pytest.raises(ValueError):
        nbd.open(nbd_server.url, "invalid")


@pytest.mark.parametrize("img_size,buf_size", [
    pytest.param(4096, 4096, id="full"),
    pytest.param(4096, 8192, id="short-aligned"),
    pytest.param(4097, 8192, id="short-unaligned"),
])
def test_readinto(nbd_server, img_size, buf_size):
    with open(nbd_server.image, "wb") as f:
        f.write(b"x" * img_size)

    nbd_server.start()

    buf = bytearray(buf_size)
    with nbd.open(nbd_server.url, "r+") as f:
        # When image size is not aligned to 512 bytes, the last read is
        # extended to the next multiple of 512.
        effective_size = f.size()

        n = f.readinto(buf)
        assert n == effective_size
        assert f.tell() == effective_size
        assert buf[:img_size] == b"x" * img_size
        assert buf[img_size:] == b"\0" * (buf_size - img_size)


@pytest.mark.parametrize("end_offset", [
    pytest.param(0, id="at-end"),
    pytest.param(1, id="after-end"),
])
def test_readinto_end(nbd_server, end_offset):
    nbd_server.start()
    buf = bytearray(512)
    with nbd.open(nbd_server.url, "r+") as f:
        offset = f.size() + end_offset
        f.seek(offset)
        n = f.readinto(buf)
        assert n == 0
        assert f.tell() == offset


@pytest.mark.parametrize("sparse", [True, False])
def test_zero_middle(nbd_server, sparse):
    nbd_server.start()
    with nbd.open(nbd_server.url, "r+", sparse=sparse) as b:
        b.write(b"xxxxxxxxxxxx")
        b.seek(4)
        assert b.zero(4) == 4

        with closing(util.aligned_buffer(12)) as buf:
            b.seek(0)
            assert b.readinto(buf) == 12
            assert buf[:] == b"xxxx\x00\x00\x00\x00xxxx"


@pytest.mark.parametrize("sparse", [True, False])
def test_zero_sparse(nbd_server, user_file, sparse):
    size = 10 * 1024**2
    qemu_img.create(user_file.path, "raw", size=size)
    nbd_server.image = user_file.path
    nbd_server.start()

    with nbd.open(nbd_server.url, "r+", sparse=sparse) as b:
        b.zero(b.size())
        b.flush()
        actual_size = os.stat(user_file.path).st_blocks * 512
        assert actual_size == 0 if sparse else b.size()


def test_close(nbd_server):
    nbd_server.start()
    with nbd.open(nbd_server.url, "r+") as b:
        pass

    # Closing twice does not do anything.
    b.close()

    # But other operations should fail.
    error = "Operation on closed backend"

    with pytest.raises(ValueError) as e:
        b.write("more")
    assert str(e.value) == error

    with pytest.raises(ValueError) as e:
        with closing(util.aligned_buffer(100)) as buf:
            b.readinto(buf)
    assert str(e.value) == error


def test_context_manager(nbd_server):
    nbd_server.start()
    with nbd.open(nbd_server.url, "r+") as b:
        b.write(b"data")
    with pytest.raises(ValueError):
        b.write("more")


def test_dirty(nbd_server):
    nbd_server.start()
    with nbd.open(nbd_server.url, "r+") as b:
        # backend created clean
        assert not b.dirty

        # write and zero dirty the backend
        b.write(b"01234")
        assert b.dirty

        b.flush()
        assert not b.dirty

        b.zero(5)
        assert b.dirty

        b.flush()
        assert not b.dirty

        # readinto, seek do not affect dirty.
        b.seek(0)
        assert not b.dirty

        with closing(util.aligned_buffer(10)) as buf:
            b.readinto(buf)
        assert not b.dirty


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_size(nbd_server, fmt):
    size = 150 * 1024**2
    nbd_server.fmt = fmt
    qemu_img.create(nbd_server.image, fmt, size=size)
    nbd_server.start()
    with nbd.open(nbd_server.url) as b:
        assert b.size() == size


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_extents_zero(nbd_server, user_file, fmt):
    size = 6 * 1024**3
    qemu_img.create(user_file.path, fmt, size=size)

    nbd_server.image = user_file.path
    nbd_server.fmt = fmt
    nbd_server.start()

    with nbd.open(nbd_server.url, "r+") as b:
        # qcow2 extents resolution is cluster size.
        data = b"x" * 64 * 1024
        b.write(data)

        # The second extent length is bigger than NBD maximum length, testing
        # that our extent length is not limited by NBD limits. The backend
        # sends multiple block status commands and merge the returned extents.
        b.seek(5 * 1024**3)
        b.write(data)

        # Holes can be reported only for qcow2 images.
        hole = fmt == "qcow2"

        assert list(b.extents()) == [
            extent.ZeroExtent(0, len(data), False, False),
            extent.ZeroExtent(len(data), 5 * 1024**3 - len(data), True, hole),
            extent.ZeroExtent(5 * 1024**3, len(data), False, False),
            extent.ZeroExtent(
                5 * 1024**3 + len(data), 1024**3 - len(data), True, hole),
        ]


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_extents_dirty_not_availabe(nbd_server, fmt):
    qemu_img.create(nbd_server.image, fmt, 65536)
    nbd_server.fmt = fmt
    nbd_server.start()

    with nbd.open(nbd_server.url, "r+", dirty=True) as b:
        with pytest.raises(errors.UnsupportedOperation):
            list(b.extents(context="dirty"))


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_clone(nbd_server, fmt):
    qemu_img.create(nbd_server.image, fmt, 65536)
    nbd_server.fmt = fmt
    nbd_server.start()

    with nbd.open(nbd_server.url, "r+") as a, \
            a.clone() as b:
        # Backends are indentical when created.
        assert a.size() == b.size()
        assert a.tell() == b.tell()
        assert a.block_size == b.block_size

        # Modifying one backend does not change the other.
        a.write(b"x" * 4096)
        assert b.tell() == 0

        # Both backends expoose the same content.
        buf = bytearray(4096)
        b.readinto(buf)
        assert buf == b"x" * 4096
